# This code is part of Qiskit.
#
# (C) Copyright IBM 2021, 2024.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

# pylint: disable=missing-class-docstring,missing-function-docstring
# pylint: disable=missing-module-docstring

import math

from test import combine

from ddt import ddt, data

from numpy.testing import assert_array_max_ulp
from qiskit.circuit import QuantumCircuit, ClassicalRegister, QuantumRegister
from qiskit.circuit.library.standard_gates import (
    CXGate,
    ECRGate,
)
from qiskit.compiler import transpile
from qiskit.providers.basic_provider import BasicSimulator
from qiskit.providers.fake_provider import GenericBackendV2
from qiskit.pulse import channels
from qiskit.quantum_info import Operator
from qiskit.transpiler import InstructionProperties
from test import QiskitTestCase  # pylint: disable=wrong-import-order

from ..legacy_cmaps import BOGOTA_CMAP, TENERIFE_CMAP
from .fake_mumbai_v2 import FakeMumbaiFractionalCX


@ddt
class TestBackendV2(QiskitTestCase):
    def setUp(self):
        super().setUp()
        self.backend = GenericBackendV2(num_qubits=2, basis_gates=["rx", "u"], seed=42)
        cx_props = {
            (0, 1): InstructionProperties(duration=5.23e-7, error=0.00098115),
        }
        self.backend.target.add_instruction(CXGate(), cx_props)
        ecr_props = {
            (1, 0): InstructionProperties(duration=4.52e-9, error=0.0000132115),
        }
        self.backend.target.add_instruction(ECRGate(), ecr_props)
        self.backend.options.set_validator("shots", (1, 4096))

    def assertMatchesTargetConstraints(self, tqc, target):
        qubit_indices = {qubit: index for index, qubit in enumerate(tqc.qubits)}
        for instruction in tqc.data:
            qubits = tuple(qubit_indices[x] for x in instruction.qubits)
            target_set = target[instruction.operation.name].keys()
            self.assertIn(
                qubits,
                target_set,
                f"qargs: {qubits} not found in target for operation {instruction.operation.name}:"
                f" {set(target_set)}",
            )

    def test_qubit_properties(self):
        """Test that qubit properties are returned as expected."""
        props = self.backend.qubit_properties([1, 0])
        assert_array_max_ulp([0.0001697368029059364, 0.00017739560485559633], [x.t1 for x in props])
        assert_array_max_ulp(
            [0.00010941773478876496, 0.00014388784397520525], [x.t2 for x in props]
        )
        assert_array_max_ulp([5487811175.818378, 5429298959.955691], [x.frequency for x in props])

    def test_legacy_qubit_properties(self):
        """Test that qubit props work for backends not using properties in target."""

        class FakeBackendV2LegacyQubitProps(GenericBackendV2):
            """Fake backend that doesn't use qubit properties via the target."""

            def qubit_properties(self, qubit):
                if isinstance(qubit, int):
                    return self.target.qubit_properties[qubit]
                return [self.target.qubit_properties[i] for i in qubit]

        props = FakeBackendV2LegacyQubitProps(num_qubits=2, seed=42).qubit_properties([1, 0])
        assert_array_max_ulp([0.0001697368029059364, 0.00017739560485559633], [x.t1 for x in props])
        assert_array_max_ulp(
            [0.00010941773478876496, 0.00014388784397520525], [x.t2 for x in props]
        )
        assert_array_max_ulp([5487811175.818378, 5429298959.955691], [x.frequency for x in props])

    def test_no_qubit_properties_raises(self):
        """Ensure that if no qubit properties are defined we raise correctly."""
        with self.assertRaises(NotImplementedError):
            BasicSimulator().qubit_properties(0)

    def test_option_bounds(self):
        """Test that option bounds are enforced."""
        with self.assertRaises(ValueError) as cm:
            self.backend.set_options(shots=8192)
        self.assertEqual(
            str(cm.exception),
            "Specified value for 'shots' is not a valid value, must be >=1 or <=4096",
        )

    @data(0, 1, 2, 3)
    def test_transpile(self, opt_level):
        """Test that transpile() works with a BackendV2 backend."""
        qc = QuantumCircuit(2)
        qc.h(1)
        qc.cz(1, 0)
        tqc = transpile(qc, self.backend, optimization_level=opt_level)
        self.assertTrue(Operator.from_circuit(tqc).equiv(qc))
        self.assertMatchesTargetConstraints(tqc, self.backend.target)

    @combine(
        opt_level=[0, 1, 2, 3],
        gate=["cx", "ecr", "cz"],
        bidirectional=[True, False],
        dsc=(
            "Test GHZ circuit with {gate} using opt level {opt_level} on backend "
            "with bidirectional={bidirectional}"
        ),
        name="{gate}_level_{opt_level}_bidirectional_{bidirectional}",
    )
    def test_5q_ghz(self, opt_level, gate, bidirectional):
        if bidirectional:
            backend = GenericBackendV2(num_qubits=5, seed=42)
        else:
            backend = GenericBackendV2(num_qubits=5, coupling_map=TENERIFE_CMAP, seed=42)
        qc = QuantumCircuit(5)
        qc.h(0)
        getattr(qc, gate)(0, 1)
        getattr(qc, gate)(2, 1)
        getattr(qc, gate)(2, 3)
        getattr(qc, gate)(4, 3)
        tqc = transpile(qc, backend, optimization_level=opt_level)
        t_op = Operator.from_circuit(tqc)
        self.assertTrue(t_op.equiv(qc))
        self.assertMatchesTargetConstraints(tqc, backend.target)

    def test_transpile_respects_arg_constraints(self):
        """Test that transpile() respects a heterogenous basis."""
        # Test CX on wrong link
        qc = QuantumCircuit(2)
        qc.h(0)
        qc.cx(1, 0)
        tqc = transpile(qc, self.backend, optimization_level=1)
        self.assertTrue(Operator.from_circuit(tqc).equiv(qc))
        # Below is done to check we're decomposing cx(1, 0) with extra
        # rotations to correct for direction. However because of fp
        # differences between windows and other platforms the optimization
        # from the 1q optimization passes differ and the output gates
        # change (while still being equivalent). This relaxes the check to
        # still ensure it's valid but not so specific that it fails on windows
        self.assertEqual(tqc.count_ops().keys(), {"cx", "u"})
        self.assertEqual(tqc.count_ops()["cx"], 1)
        self.assertLessEqual(tqc.count_ops()["u"], 4)
        self.assertMatchesTargetConstraints(tqc, self.backend.target)
        # Test ECR on wrong link
        qc = QuantumCircuit(2)
        qc.h(0)
        qc.ecr(0, 1)
        tqc = transpile(qc, self.backend, optimization_level=1)
        self.assertTrue(Operator.from_circuit(tqc).equiv(qc))
        self.assertEqual(tqc.count_ops(), {"ecr": 1, "u": 4})
        self.assertMatchesTargetConstraints(tqc, self.backend.target)

    def test_transpile_relies_on_gate_direction(self):
        """Test that transpile() relies on gate direction pass for 2q."""
        qc = QuantumCircuit(2)
        qc.h(0)
        qc.ecr(0, 1)
        tqc = transpile(qc, self.backend, optimization_level=1)
        expected = QuantumCircuit(2)
        expected.u(0, 0, -math.pi, 0)
        expected.u(math.pi / 2, 0, 0, 1)
        expected.ecr(1, 0)
        expected.u(math.pi / 2, 0, -math.pi, 0)
        expected.u(math.pi / 2, 0, -math.pi, 1)
        self.assertTrue(Operator.from_circuit(tqc).equiv(qc))
        self.assertEqual(tqc.count_ops(), {"ecr": 1, "u": 4})
        self.assertMatchesTargetConstraints(tqc, self.backend.target)

    def test_transpile_mumbai_target(self):
        """Test that transpile respects a more involved target for a fake mumbai."""
        backend = FakeMumbaiFractionalCX()
        qc = QuantumCircuit(2)
        qc.h(0)
        qc.cx(1, 0)
        qc.measure_all()
        tqc = transpile(qc, backend, optimization_level=1)
        qr = QuantumRegister(27, "q")
        cr = ClassicalRegister(2, "meas")
        expected = QuantumCircuit(qr, cr, global_phase=math.pi / 4)
        expected.rz(math.pi / 2, 0)
        expected.sx(0)
        expected.rz(math.pi / 2, 0)
        expected.cx(1, 0)
        expected.barrier(qr[0], qr[1])
        expected.measure(qr[0], cr[0])
        expected.measure(qr[1], cr[1])
        self.assertEqual(expected, tqc)

    @data(0, 1, 2, 3, 4)
    def test_drive_channel(self, qubit):
        """Test getting drive channel with qubit index."""
        backend = GenericBackendV2(num_qubits=5, seed=42)
        with self.assertWarns(DeprecationWarning):
            chan = backend.drive_channel(qubit)
            ref = channels.DriveChannel(qubit)
        self.assertEqual(chan, ref)

    @data(0, 1, 2, 3, 4)
    def test_measure_channel(self, qubit):
        """Test getting measure channel with qubit index."""
        backend = GenericBackendV2(num_qubits=5, seed=42)
        with self.assertWarns(DeprecationWarning):
            chan = backend.measure_channel(qubit)
            ref = channels.MeasureChannel(qubit)
        self.assertEqual(chan, ref)

    @data(0, 1, 2, 3, 4)
    def test_acquire_channel(self, qubit):
        """Test getting acquire channel with qubit index."""
        backend = GenericBackendV2(num_qubits=5, seed=42)
        with self.assertWarns(DeprecationWarning):
            chan = backend.acquire_channel(qubit)
            ref = channels.AcquireChannel(qubit)
        self.assertEqual(chan, ref)

    @data((4, 3), (3, 4), (3, 2), (2, 3), (1, 2), (2, 1), (1, 0), (0, 1))
    def test_control_channel(self, qubits):
        """Test getting acquire channel with qubit index."""
        bogota_cr_channels_map = {
            (4, 3): 7,
            (3, 4): 6,
            (3, 2): 5,
            (2, 3): 4,
            (1, 2): 2,
            (2, 1): 3,
            (1, 0): 1,
            (0, 1): 0,
        }
        backend = GenericBackendV2(num_qubits=5, coupling_map=BOGOTA_CMAP, seed=42)
        with self.assertWarns(DeprecationWarning):
            chan = backend.control_channel(qubits)[0]
            ref = channels.ControlChannel(bogota_cr_channels_map[qubits])
        self.assertEqual(chan, ref)



# This code is part of Qiskit.
#
# (C) Copyright IBM 2020.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Result object to analyse image comparisons"""

import os
import json
from PIL import Image, ImageChops, ImageDraw

SWD = os.path.dirname(os.path.abspath(__file__))


class Results:
    """Result object to analyse image comparisons"""

    def __init__(self, names, directory):
        self.names = names
        self.directory = directory
        self.data = {}
        self.exact_match = []
        self.mismatch = []
        self.missing = []
        datafilename = os.path.join(SWD, directory, "result_test.json")
        if os.path.exists(datafilename):
            with open(datafilename) as datafile:
                self.data = json.load(datafile)

    @staticmethod
    def _black_or_b(diff_image, image, reference, opacity=0.85):
        """Copied from https://stackoverflow.com/a/30307875"""
        thresholded_diff = diff_image
        for _ in range(3):
            thresholded_diff = ImageChops.add(thresholded_diff, thresholded_diff)
        size = diff_image.size
        mask = Results._new_gray(size, int(255 * (opacity)))
        shade = Results._new_gray(size, 0)
        new = reference.copy()
        new.paste(shade, mask=mask)
        if image.size != new.size:
            image = image.resize(new.size)
        if image.size != thresholded_diff.size:
            thresholded_diff = thresholded_diff.resize(image.size)
        new.paste(image, mask=thresholded_diff)
        return new

    @staticmethod
    def _get_black_pixels(image):
        black_and_white_version = image.convert("1")
        black_pixels = black_and_white_version.histogram()[0]
        return black_pixels

    @staticmethod
    def _similarity_ratio(current, expected):
        diff_name = current.split(".")
        diff_name.insert(-1, "diff")
        diff_name = ".".join(diff_name)

        current = Image.open(current)
        expected = Image.open(expected)

        diff = ImageChops.difference(expected, current).convert("L")
        Results._black_or_b(diff, current, expected).save(diff_name, "PNG")
        black_pixels = Results._get_black_pixels(diff)
        total_pixels = diff.size[0] * diff.size[1]
        return black_pixels / total_pixels, diff_name

    @staticmethod
    def _new_gray(size, color):
        img = Image.new("L", size)
        drawing = ImageDraw.Draw(img)
        drawing.rectangle((0, 0) + size, color)
        return img

    @staticmethod
    def passed_result_html(result, reference, diff, title):
        """Creates the html for passing tests"""
        ret = f'<details><summary style="background-color:lightgreen;"> {title} </summary>'
        ret += "<table>"
        ret += f'<tr><td><img src="{result}"</td>'
        ret += f'<td><img src="{reference}"</td>'
        ret += f'<td><img src="{diff}"</td>'
        ret += "</tr></table></details>"
        return ret

    @staticmethod
    def failed_result_html(result, reference, diff, title):
        """Creates the html for failing tests"""
        ret = f'<details open><summary style="background-color:lightcoral;"> {title} </summary>'
        ret += "<table>"
        ret += f'<tr><td><img src="{result}"</td>'
        ret += f'<td><img src="{reference}"</td>'
        ret += f'<td><img src="{diff}"</td>'
        ret += "</tr></table></details>"
        return ret

    @staticmethod
    def no_reference_html(result, title):
        """Creates the html for missing-reference tests"""
        ret = f'<details><summary style="background-color:lightgrey;"> {title} </summary>'
        ret += f'<table><tr><td><img src="{result}"</td>'
        ret += "</tr></table></details>"
        return ret

    def diff_images(self):
        """Creates the table with the image comparison"""
        for name in self.names:
            ratio = diff_name = title = None
            fullpath_name = os.path.join(self.directory, name)
            fullpath_reference = os.path.join(self.directory, "references", name)

            if os.path.exists(os.path.join(SWD, fullpath_reference)):
                ratio, diff_name = Results._similarity_ratio(fullpath_name, fullpath_reference)
                title = f"<tt><b>{name}</b> | {self.data[name]['testname']} </tt> | ratio: {ratio}"
                if ratio == 1:
                    self.exact_match.append(fullpath_name)
                else:
                    self.mismatch.append(fullpath_name)
            else:
                self.missing.append(fullpath_name)

            self.data[name]["ratio"] = ratio
            self.data[name]["diff_name"] = diff_name
            self.data[name]["title"] = title

    def _repr_html_(self):
        ret = self.summary()
        ret += "<div>"
        for name in self.names:
            fullpath_name = os.path.join(self.directory, name)
            fullpath_reference = os.path.join(self.directory, "references", name)
            if os.path.exists(os.path.join(SWD, fullpath_reference)):
                if self.data[name]["ratio"] == 1:
                    ret += Results.passed_result_html(
                        fullpath_name,
                        fullpath_reference,
                        self.data[name]["diff_name"],
                        self.data[name]["title"],
                    )
                else:
                    ret += Results.failed_result_html(
                        fullpath_name,
                        fullpath_reference,
                        self.data[name]["diff_name"],
                        self.data[name]["title"],
                    )
            else:
                title = (
                    f'Download <a download="{name}" href="{fullpath_name}">this image</a>'
                    f" to <tt>{fullpath_reference}</tt>"
                    " and add/push to the repo</td>"
                )
                ret += Results.no_reference_html(fullpath_name, title)
        ret += "</div>"
        return ret


if __name__ == "__main__":
    # collect results for circuit tests
    RESULT_FILES_CIRCUIT = []
    for file in os.listdir(os.path.join(SWD, "mpl/circuit")):
        if file.endswith(".png") and not file.endswith(".diff.png"):
            RESULT_FILES_CIRCUIT.append(file)
    RESULTS_CIRCUIT = Results(sorted(RESULT_FILES_CIRCUIT), "mpl/circuit")
    RESULTS_CIRCUIT.diff_images()

    # collect results for graph tests
    RESULT_FILES_GRAPH = []
    for file in os.listdir(os.path.join(SWD, "mpl/graph")):
        if file.endswith(".png") and not file.endswith(".diff.png"):
            RESULT_FILES_GRAPH.append(file)
    RESULTS_GRAPH = Results(sorted(RESULT_FILES_GRAPH), "mpl/graph")
    RESULTS_GRAPH.diff_images()



# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2023.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Utilities for providers."""
from __future__ import annotations

import logging
from collections.abc import Callable

from qiskit.providers.backend import Backend

logger = logging.getLogger(__name__)


def filter_backends(
    backends: list[Backend], filters: Callable[[Backend], bool] | None = None, **kwargs
) -> list[Backend]:
    """Return the backends matching the specified filtering.

    Filter the `backends` list by their `configuration` or `status`
    attributes, or from a boolean callable. The criteria for filtering can
    be specified via `**kwargs` or as a callable via `filters`, and the
    backends must fulfill all specified conditions.

    Args:
        backends (list[Backend]): list of backends.
        filters (callable): filtering conditions as a callable.
        **kwargs: dict of criteria.

    Returns:
        list[Backend]: a list of backend instances matching the
            conditions.
    """

    def _match_all(obj, criteria):
        """Return True if all items in criteria matches items in obj."""
        return all(getattr(obj, key_, None) == value_ for key_, value_ in criteria.items())

    # Inspect the backends to decide which filters belong to
    # backend.configuration and which ones to backend.status, as it does
    # not involve querying the API.
    configuration_filters = {}
    status_filters = {}
    for key, value in kwargs.items():
        if all(key in backend.configuration() for backend in backends):
            configuration_filters[key] = value
        else:
            status_filters[key] = value

    # 1. Apply backend.configuration filtering.
    if configuration_filters:
        backends = [b for b in backends if _match_all(b.configuration(), configuration_filters)]

    # 2. Apply backend.status filtering (it involves one API call for
    # each backend).
    if status_filters:
        backends = [b for b in backends if _match_all(b.status(), status_filters)]

    # 3. Apply acceptor filter.
    backends = list(filter(filters, backends))

    return backends


def resolve_backend_name(
    name: str, backends: list[Backend], deprecated: dict[str, str], aliased: dict[str, list[str]]
) -> str:
    """Resolve backend name from a deprecated name or an alias.

    A group will be resolved in order of member priorities, depending on
    availability.

    Args:
        name (str): name of backend to resolve
        backends (list[Backend]): list of available backends.
        deprecated (dict[str: str]): dict of deprecated names.
        aliased (dict[str: list[str]]): dict of aliased names.

    Returns:
        str: resolved name (name of an available backend)

    Raises:
        LookupError: if name cannot be resolved through regular available
            names, nor deprecated, nor alias names.
    """
    # account for BackendV2
    available = []
    for backend in backends:
        available.append(backend.name() if backend.version == 1 else backend.name)

    resolved_name = deprecated.get(name, aliased.get(name, name))
    if isinstance(resolved_name, list):
        resolved_name = next((b for b in resolved_name if b in available), "")

    if resolved_name not in available:
        raise LookupError(f"backend '{name}' not found.")

    if name in deprecated:
        logger.warning("Backend '%s' is deprecated. Use '%s'.", name, resolved_name)

    return resolved_name



# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2019.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""
Configurations for pulse experiments.
"""
from __future__ import annotations
import numpy as np

from .channels import DriveChannel, MeasureChannel
from .exceptions import PulseError


def _assert_nested_dict_equal(a: dict, b: dict):
    if len(a) != len(b):
        return False
    for key in a:
        if key in b:
            if isinstance(a[key], dict):
                if not _assert_nested_dict_equal(a[key], b[key]):
                    return False
            elif isinstance(a[key], np.ndarray):
                if not np.all(a[key] == b[key]):
                    return False
            else:
                if a[key] != b[key]:
                    return False
        else:
            return False
    return True


class Kernel:
    """Settings for this Kernel, which is responsible for integrating time series (raw) data
    into IQ points.
    """

    def __init__(self, name: str | None = None, **params):
        """Create new kernel.

        Args:
            name: Name of kernel to be used
            params: Any settings for kerneling.
        """
        self.name = name
        self.params = params

    def __repr__(self):
        name_repr = "'" + self.name + "', "
        params_repr = ", ".join(f"{str(k)}={str(v)}" for k, v in self.params.items())
        return f"{self.__class__.__name__}({name_repr}{params_repr})"

    def __eq__(self, other):
        if isinstance(other, Kernel):
            return _assert_nested_dict_equal(self.__dict__, other.__dict__)
        return False


class Discriminator:
    """Setting for this Discriminator, which is responsible for classifying kerneled IQ points
    into 0/1 state results.
    """

    def __init__(self, name: str | None = None, **params):
        """Create new discriminator.

        Args:
            name: Name of discriminator to be used
            params: Any settings for discrimination.
        """
        self.name = name
        self.params = params

    def __repr__(self):
        name_repr = "'" + self.name + "', " or ""
        params_repr = ", ".join(f"{str(k)}={str(v)}" for k, v in self.params.items())
        return f"{self.__class__.__name__}({name_repr}{params_repr})"

    def __eq__(self, other):
        if isinstance(other, Discriminator):
            return _assert_nested_dict_equal(self.__dict__, other.__dict__)
        return False


class LoRange:
    """Range of LO frequency."""

    def __init__(self, lower_bound: float, upper_bound: float):
        self._lb = lower_bound
        self._ub = upper_bound

    def includes(self, lo_freq: complex) -> bool:
        """Whether `lo_freq` is within the `LoRange`.

        Args:
            lo_freq: LO frequency to be validated

        Returns:
            bool: True if lo_freq is included in this range, otherwise False
        """
        if self._lb <= abs(lo_freq) <= self._ub:
            return True
        return False

    @property
    def lower_bound(self) -> float:
        """Lower bound of the LO range"""
        return self._lb

    @property
    def upper_bound(self) -> float:
        """Upper bound of the LO range"""
        return self._ub

    def __repr__(self):
        return f"{self.__class__.__name__}({self._lb:f}, {self._ub:f})"

    def __eq__(self, other):
        """Two LO ranges are the same if they are of the same type, and
        have the same frequency range

        Args:
            other (LoRange): other LoRange

        Returns:
            bool: are self and other equal.
        """
        if type(self) is type(other) and self._ub == other._ub and self._lb == other._lb:
            return True
        return False


class LoConfig:
    """Pulse channel LO frequency container."""

    def __init__(
        self,
        channel_los: dict[DriveChannel | MeasureChannel, float] | None = None,
        lo_ranges: dict[DriveChannel | MeasureChannel, LoRange | tuple[int, int]] | None = None,
    ):
        """Lo channel configuration data structure.

        Args:
            channel_los: Dictionary of mappings from configurable channel to lo
            lo_ranges: Dictionary of mappings to be enforced from configurable channel to `LoRange`

        Raises:
            PulseError: If channel is not configurable or set lo is out of range.

        """
        self._q_lo_freq: dict[DriveChannel, float] = {}
        self._m_lo_freq: dict[MeasureChannel, float] = {}
        self._lo_ranges: dict[DriveChannel | MeasureChannel, LoRange] = {}

        lo_ranges = lo_ranges if lo_ranges else {}
        for channel, freq in lo_ranges.items():
            self.add_lo_range(channel, freq)

        channel_los = channel_los if channel_los else {}
        for channel, freq in channel_los.items():
            self.add_lo(channel, freq)

    def add_lo(self, channel: DriveChannel | MeasureChannel, freq: float):
        """Add a lo mapping for a channel."""
        if isinstance(channel, DriveChannel):
            # add qubit_lo_freq
            self.check_lo(channel, freq)
            self._q_lo_freq[channel] = freq
        elif isinstance(channel, MeasureChannel):
            # add meas_lo_freq
            self.check_lo(channel, freq)
            self._m_lo_freq[channel] = freq
        else:
            raise PulseError(f"Specified channel {channel.name} cannot be configured.")

    def add_lo_range(
        self, channel: DriveChannel | MeasureChannel, lo_range: LoRange | tuple[int, int]
    ):
        """Add lo range to configuration.

        Args:
            channel: Channel to add lo range for
            lo_range: Lo range to add

        """
        if isinstance(lo_range, (list, tuple)):
            lo_range = LoRange(*lo_range)
        self._lo_ranges[channel] = lo_range

    def check_lo(self, channel: DriveChannel | MeasureChannel, freq: float) -> bool:
        """Check that lo is valid for channel.

        Args:
            channel: Channel to validate lo for
            freq: lo frequency
        Raises:
            PulseError: If freq is outside of channels range
        Returns:
            True if lo is valid for channel
        """
        lo_ranges = self._lo_ranges
        if channel in lo_ranges:
            lo_range = lo_ranges[channel]
            if not lo_range.includes(freq):
                raise PulseError(f"Specified LO freq {freq:f} is out of range {lo_range}")
        return True

    def channel_lo(self, channel: DriveChannel | MeasureChannel) -> float:
        """Return channel lo.

        Args:
            channel: Channel to get lo for
        Raises:
            PulseError: If channel is not configured
        Returns:
            Lo of supplied channel if present
        """
        if isinstance(channel, DriveChannel):
            if channel in self.qubit_los:
                return self.qubit_los[channel]

        if isinstance(channel, MeasureChannel):
            if channel in self.meas_los:
                return self.meas_los[channel]

        raise PulseError(f"Channel {channel} is not configured")

    @property
    def qubit_los(self) -> dict[DriveChannel, float]:
        """Returns dictionary mapping qubit channels (DriveChannel) to los."""
        return self._q_lo_freq

    @property
    def meas_los(self) -> dict[MeasureChannel, float]:
        """Returns dictionary mapping measure channels (MeasureChannel) to los."""
        return self._m_lo_freq



# This code is part of Qiskit.
#
# (C) Copyright IBM 2017.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Multiprocessing utilities"""

import multiprocessing as mp
import platform
import os


def local_hardware_info():
    """Basic hardware information about the local machine.

    Gives actual number of CPU's in the machine, even when hyperthreading is
    turned on. CPU count defaults to 1 when true count can't be determined.

    Returns:
        dict: The hardware information.
    """

    if hasattr(os, "sched_getaffinity"):
        num_cpus = len(os.sched_getaffinity(0))
    else:
        num_cpus = os.cpu_count()
    if num_cpus is None:
        num_cpus = 1
    else:
        num_cpus = int(num_cpus / 2) or 1

    results = {
        "python_compiler": platform.python_compiler(),
        "python_build": ", ".join(platform.python_build()),
        "python_version": platform.python_version(),
        "os": platform.system(),
        "cpus": num_cpus,
    }
    return results


def is_main_process():
    """Checks whether the current process is the main one"""
    if platform.system() == "Windows":
        return not isinstance(mp.current_process(), mp.context.SpawnProcess)
    else:
        return not isinstance(
            mp.current_process(), (mp.context.ForkProcess, mp.context.SpawnProcess)
        )



# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2020.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Test the piecewise Chebyshev approximation."""

import unittest
from collections import defaultdict
import numpy as np
from ddt import ddt, data, unpack

from qiskit.circuit import QuantumCircuit
from qiskit.circuit.library.arithmetic.piecewise_chebyshev import PiecewiseChebyshev
from qiskit.quantum_info import Statevector
from test import QiskitTestCase  # pylint: disable=wrong-import-order


@ddt
class TestPiecewiseChebyshev(QiskitTestCase):
    """Test the piecewise Chebyshev approximation."""

    def assertFunctionIsCorrect(self, function_circuit, reference):
        """Assert that ``function_circuit`` implements the reference function ``reference``."""
        function_circuit._build()
        num_state_qubits = function_circuit.num_state_qubits
        num_ancilla_qubits = function_circuit.num_ancillas
        circuit = QuantumCircuit(num_state_qubits + 1 + num_ancilla_qubits)
        circuit.h(list(range(num_state_qubits)))
        circuit.append(function_circuit.to_instruction(), list(range(circuit.num_qubits)))
        statevector = Statevector(circuit)
        probabilities = defaultdict(float)
        for i, statevector_amplitude in enumerate(statevector):
            i = bin(i)[2:].zfill(circuit.num_qubits)[num_ancilla_qubits:]
            probabilities[i] += np.real(np.abs(statevector_amplitude) ** 2)

        unrolled_probabilities = []
        unrolled_expectations = []
        for i, probability in probabilities.items():
            x, last_qubit = int(i[1:], 2), i[0]
            if last_qubit == "0":
                expected_amplitude = np.cos(reference(x)) / np.sqrt(2**num_state_qubits)
            else:
                expected_amplitude = np.sin(reference(x)) / np.sqrt(2**num_state_qubits)

            unrolled_probabilities += [probability]
            unrolled_expectations += [np.real(np.abs(expected_amplitude) ** 2)]

        np.testing.assert_array_almost_equal(
            unrolled_probabilities, unrolled_expectations, decimal=1
        )

    @data(
        (lambda x: np.arcsin(1 / x), 2, [2, 4], 2),
        (lambda x: x / 8, 1, [1, 8], 3),
        (np.sqrt, 2, None, 2),
    )
    @unpack
    def test_piecewise_chebyshev(self, f_x, degree, breakpoints, num_state_qubits):
        """Test the piecewise Chebyshev approximation."""

        def pw_poly(x):
            if breakpoints:
                if len(breakpoints) > 1:
                    start = breakpoints[0]
                    end = breakpoints[-1]
                else:
                    start = breakpoints[0]
                    end = 2**num_state_qubits
            else:
                start = 0
                end = 2**num_state_qubits
            if start <= x < end:
                return f_x(x)
            return np.arcsin(1)

        pw_approximation = PiecewiseChebyshev(f_x, degree, breakpoints, num_state_qubits)

        self.assertFunctionIsCorrect(pw_approximation, pw_poly)

    def test_piecewise_chebyshev_mutability(self):
        """Test the mutability of the piecewise Chebyshev approximation."""

        def pw_poly(x, f_x):
            if breakpoints[0] <= x < breakpoints[-1]:
                return f_x(x)
            return np.arcsin(1)

        def f_x_1(x):
            return x / 2

        pw_approximation = PiecewiseChebyshev(f_x_1)

        with self.subTest(msg="missing number of state qubits"):
            with self.assertRaises(AttributeError):  # no state qubits set
                _ = str(pw_approximation.draw())

        with self.subTest(msg="default setup, just setting number of state qubits"):
            pw_approximation.num_state_qubits = 2
            pw_approximation.f_x = f_x_1
            # set to the default breakpoints for pw_poly
            breakpoints = [0, 4]
            pw_approximation.breakpoints = breakpoints
            self.assertFunctionIsCorrect(pw_approximation, lambda x: pw_poly(x, f_x_1))

        def f_x_2(x):
            return x / 4

        with self.subTest(msg="setting non-default values"):
            breakpoints = [0, 2]
            degree = 2
            pw_approximation.breakpoints = breakpoints
            pw_approximation.degree = degree
            pw_approximation.f_x = f_x_2
            self.assertFunctionIsCorrect(pw_approximation, lambda x: pw_poly(x, f_x_2))

        def f_x_3(x):
            return x**2

        with self.subTest(msg="changing all values"):
            pw_approximation.num_state_qubits = 4
            breakpoints = [1, 3, 6]
            degree = 3
            pw_approximation.breakpoints = breakpoints
            pw_approximation.degree = degree
            pw_approximation.f_x = f_x_3
            self.assertFunctionIsCorrect(pw_approximation, lambda x: pw_poly(x, f_x_3))


if __name__ == "__main__":
    unittest.main()



# This code is part of Qiskit.
#
# (C) Copyright IBM 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Test for unit conversion functions."""

from ddt import ddt, data

from qiskit.utils import apply_prefix, detach_prefix
from test import QiskitTestCase  # pylint: disable=wrong-import-order


@ddt
class TestUnitConversion(QiskitTestCase):
    """Test the unit conversion utilities."""

    def test_apply_prefix(self):
        """Test applying prefix to value."""
        ref_values = [
            ([1.0, "THz"], 1e12),
            ([1.0, "GHz"], 1e9),
            ([1.0, "MHz"], 1e6),
            ([1.0, "kHz"], 1e3),
            ([1.0, "mHz"], 1e-3),
            ([1.0, "µHz"], 1e-6),
            ([1.0, "uHz"], 1e-6),
            ([1.0, "nHz"], 1e-9),
            ([1.0, "pHz"], 1e-12),
        ]

        for args, ref_ret in ref_values:
            self.assertEqual(apply_prefix(*args), ref_ret)

    def test_not_convert_meter(self):
        """Test not apply prefix to meter."""
        self.assertEqual(apply_prefix(1.0, "m"), 1.0)

    def test_detach_prefix(self):
        """Test detach prefix from the value."""
        ref_values = [
            (1e12, (1.0, "T")),
            (1e11, (100.0, "G")),
            (1e10, (10.0, "G")),
            (1e9, (1.0, "G")),
            (1e8, (100.0, "M")),
            (1e7, (10.0, "M")),
            (1e6, (1.0, "M")),
            (1e5, (100.0, "k")),
            (1e4, (10.0, "k")),
            (1e3, (1.0, "k")),
            (100, (100.0, "")),
            (10, (10.0, "")),
            (1.0, (1.0, "")),
            (0.1, (100.0, "m")),
            (0.01, (10.0, "m")),
            (1e-3, (1.0, "m")),
            (1e-4, (100.0, "µ")),
            (1e-5, (10.0, "µ")),
            (1e-6, (1.0, "µ")),
            (1e-7, (100.0, "n")),
            (1e-8, (10.0, "n")),
            (1e-9, (1.0, "n")),
            (1e-10, (100.0, "p")),
            (1e-11, (10.0, "p")),
            (1e-12, (1.0, "p")),
        ]

        for arg, ref_rets in ref_values:
            self.assertTupleEqual(detach_prefix(arg), ref_rets)

    def test_detach_prefix_with_zero(self):
        """Test detach prefix by input zero."""
        self.assertTupleEqual(detach_prefix(0.0), (0.0, ""))

    def test_detach_prefix_with_negative(self):
        """Test detach prefix by input negative values."""
        self.assertTupleEqual(detach_prefix(-1.234e7), (-12.34, "M"))

    def test_detach_prefix_with_value_too_large(self):
        """Test detach prefix by input too large value."""
        with self.assertRaises(Exception):
            self.assertTupleEqual(detach_prefix(1e20), (1e20, ""))

    def test_detach_prefix_with_value_too_small(self):
        """Test detach prefix by input too small value."""
        with self.assertRaises(Exception):
            self.assertTupleEqual(detach_prefix(1e-20), (1e-20, ""))

    def test_rounding(self):
        """Test detach prefix with decimal specification."""
        ret = detach_prefix(999_999.991)
        self.assertTupleEqual(ret, (999.999991, "k"))

        ret = detach_prefix(999_999.991, decimal=4)
        self.assertTupleEqual(ret, (1.0, "M"))

        ret = detach_prefix(999_999.991, decimal=5)
        self.assertTupleEqual(ret, (999.99999, "k"))

    @data(
        -20.791378538739863,
        9.242757760406565,
        2.7366806276451543,
        9.183776167253349,
        7.658091886606501,
        -12.21553566621071,
        8.914055281578145,
        1.2518807770035825,
        -6.652899195646036,
        -4.647159596697976,
    )
    def test_get_same_value_after_attach_detach(self, value: float):
        """Test if same value can be obtained."""
        unit = "Hz"

        for prefix in ["P", "T", "G", "k", "m", "µ", "n", "p", "f"]:
            scaled_val = apply_prefix(value, prefix + unit)
            test_val, ret_prefix = detach_prefix(scaled_val)
            self.assertAlmostEqual(test_val, value)
            self.assertEqual(prefix, ret_prefix)

    def test_get_symbol_mu(self):
        """Test if µ is returned rather than u."""
        _, prefix = detach_prefix(3e-6)
        self.assertEqual(prefix, "µ")



# This code is part of Qiskit.
#
# (C) Copyright IBM 2024.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Unit tests for BitArray."""

from itertools import product
from test import QiskitTestCase

import ddt
import numpy as np

from qiskit.primitives.containers import BitArray
from qiskit.quantum_info import Pauli, SparsePauliOp
from qiskit.result import Counts


def u_8(arr):
    """Convenience function to instantiate a ``numpy.uint8`` array with few characters."""
    return np.array(arr, dtype=np.uint8)


@ddt.ddt
class BitArrayTestCase(QiskitTestCase):
    """Test the DataBin class."""

    @ddt.idata(product([(), (3, 4, 5)], [1, 6], [8, 13, 26]))
    @ddt.unpack
    def test_container(self, shape, num_shots, num_bits):
        """Test the constructor and basic attributes."""
        num_bytes = num_bits // 8 + (num_bits % 8 > 0)
        size = np.prod(shape).astype(int) * num_shots * num_bytes
        arr = np.arange(size, dtype=np.uint8).reshape(shape + (num_shots, num_bytes))

        bit_array = BitArray(arr, num_bits)
        self.assertEqual(bit_array.shape, shape)
        self.assertEqual(bit_array.size, np.prod(shape).astype(int))
        self.assertEqual(bit_array.ndim, len(shape))
        self.assertEqual(bit_array.num_shots, num_shots)
        self.assertEqual(bit_array.num_bits, num_bits)
        self.assertTrue(np.all(bit_array.array == arr))
        self.assertEqual(bit_array.array.shape[-1], num_bytes)

    def test_constructor_exceptions(self):
        """Test the constructor raises exceptions properly."""
        with self.assertRaisesRegex(TypeError, "must be a numpy.ndarray"):
            BitArray([], 1)

        with self.assertRaisesRegex(TypeError, "must have dtype uint8"):
            BitArray(np.zeros(1, dtype=float), 1)

        with self.assertRaisesRegex(ValueError, "at least two axes"):
            BitArray(np.zeros((1,), dtype=np.uint8), 1)

        with self.assertRaisesRegex(ValueError, "3 bytes per shot"):
            BitArray(np.empty((2, 3, 4, 5), dtype=np.uint8), 23)

    def test_get_counts(self):
        """Test conversion to counts."""
        # note that [234, 100] requires 16 bits, not 15; we are testing that get_counts ignores the
        # junk columns
        bit_array = BitArray(u_8([[3, 5], [3, 5], [234, 100]]), num_bits=15)
        bs1 = "0000011" + "00000101"  # 3, 5
        bs2 = "1101010" + "01100100"  # 234, 100
        self.assertEqual(bit_array.get_counts(), {bs1: 2, bs2: 1})

        bit_array = BitArray(
            u_8([[[3, 5], [3, 5], [234, 100]], [[0, 1], [1, 0], [1, 0]]]),
            num_bits=15,
        )
        bs1 = "0000011" + "00000101"  # 3, 5
        bs2 = "1101010" + "01100100"  # 234, 100
        self.assertEqual(bit_array.get_counts(0), {bs1: 2, bs2: 1})
        bs3 = "0000000" + "00000001"  # 0, 1
        bs4 = "0000001" + "00000000"  # 1, 0
        self.assertEqual(bit_array.get_counts(1), {bs3: 1, bs4: 2})

        # test that providing no location takes the union over all shots
        self.assertEqual(bit_array.get_counts(), {bs1: 2, bs2: 1, bs3: 1, bs4: 2})

    def test_get_int_counts(self):
        """Test conversion to int counts."""
        # note that [234, 100] requires 16 bits, not 15; we are testing that get_counts ignores the
        # junk columns
        bit_array = BitArray(u_8([[3, 5], [3, 5], [234, 100]]), num_bits=15)
        val1 = (3 << 8) + 5
        val2 = ((234 & 127) << 8) + 100
        self.assertEqual(bit_array.get_int_counts(), {val1: 2, val2: 1})

        bit_array = BitArray(
            u_8([[[3, 5], [3, 5], [234, 100]], [[0, 1], [1, 0], [1, 0]]]),
            num_bits=15,
        )
        val1 = (3 << 8) + 5
        val2 = ((234 & 127) << 8) + 100
        self.assertEqual(bit_array.get_int_counts(0), {val1: 2, val2: 1})
        val3 = 1
        val4 = 1 << 8
        self.assertEqual(bit_array.get_int_counts(1), {val3: 1, val4: 2})

        # test that providing no location takes the union over all shots
        self.assertEqual(bit_array.get_int_counts(), {val1: 2, val2: 1, val3: 1, val4: 2})

    def test_get_bitstrings(self):
        """Test conversion to bitstrings."""
        # note that [234, 100] requires 16 bits, not 15; we are testing that get_counts ignores the
        # junk columns
        bit_array = BitArray(u_8([[3, 5], [3, 5], [234, 100]]), num_bits=15)
        bs1 = "0000011" + "00000101"  # 3, 5
        bs2 = "1101010" + "01100100"  # 234, 100
        self.assertEqual(bit_array.get_bitstrings(), [bs1, bs1, bs2])

        bit_array = BitArray(
            u_8([[[3, 5], [3, 5], [234, 100]], [[0, 1], [1, 0], [1, 0]]]),
            num_bits=15,
        )
        bs1 = "0000011" + "00000101"  # 3, 5
        bs2 = "1101010" + "01100100"  # 234, 100
        self.assertEqual(bit_array.get_bitstrings(0), [bs1, bs1, bs2])
        bs3 = "0000000" + "00000001"  # 0, 1
        bs4 = "0000001" + "00000000"  # 1, 0
        self.assertEqual(bit_array.get_bitstrings(1), [bs3, bs4, bs4])

        # test that providing no location takes the union over all shots
        self.assertEqual(bit_array.get_bitstrings(), [bs1, bs1, bs2, bs3, bs4, bs4])

    def test_equality(self):
        """Test the equality operator"""
        ba1 = BitArray.from_bool_array([[1, 0, 0], [1, 1, 0]])
        ba2 = BitArray.from_bool_array([[1, 0, 0], [1, 1, 0]])
        ba3 = BitArray.from_bool_array([[1, 1, 0], [1, 1, 0]])
        ba4 = BitArray.from_bool_array([[[1, 0, 0], [1, 1, 0]]])
        self.assertEqual(ba1, ba1)
        self.assertEqual(ba1, ba2)
        self.assertNotEqual(ba1, ba3)
        self.assertNotEqual(ba1, ba4)

        ba5 = BitArray(u_8([[4, 200], [255, 10]]), num_bits=13)
        ba6 = BitArray(u_8([[4, 200], [255, 10]]), num_bits=12)
        ba7 = BitArray(u_8([[4, 200], [31, 10]]), num_bits=13)
        self.assertNotEqual(ba5, ba6)
        self.assertEqual(ba5, ba7)  # test masking

    def test_logical_and(self):
        """Test the logical AND operator."""
        ba1 = BitArray.from_bool_array([[1, 0, 0], [1, 1, 0]])
        ba2 = BitArray.from_bool_array([[1, 1, 1], [0, 1, 1]])
        self.assertEqual(ba1 & ba2, BitArray.from_bool_array([[1, 0, 0], [0, 1, 0]]))

        ba1 = BitArray.from_bool_array([[1, 0, 0], [1, 1, 0]])
        ba2 = BitArray.from_bool_array([[1, 1, 0]])
        self.assertEqual(ba1 & ba2, BitArray.from_bool_array([[1, 0, 0], [1, 1, 0]]))

    def test_logical_or(self):
        """Test the logical OR operator."""
        ba1 = BitArray.from_bool_array([[1, 0, 0], [1, 1, 0]])
        ba2 = BitArray.from_bool_array([[1, 0, 1], [0, 1, 0]])
        self.assertEqual(ba1 | ba2, BitArray.from_bool_array([[1, 0, 1], [1, 1, 0]]))

        ba1 = BitArray.from_bool_array([[1, 0, 0], [1, 1, 0]])
        ba2 = BitArray.from_bool_array([[1, 1, 0]])
        self.assertEqual(ba1 | ba2, BitArray.from_bool_array([[1, 1, 0], [1, 1, 0]]))

    def test_logical_not(self):
        """Test the logical OR operator."""
        ba = BitArray.from_bool_array([[1, 0, 0], [1, 1, 0]])
        self.assertEqual(~ba, BitArray.from_bool_array([[0, 1, 1], [0, 0, 1]]))

    def test_logical_xor(self):
        """Test the logical XOR operator."""
        ba1 = BitArray.from_bool_array([[1, 0, 0], [1, 1, 0]])
        ba2 = BitArray.from_bool_array([[1, 0, 1], [0, 1, 0]])
        self.assertEqual(ba1 ^ ba2, BitArray.from_bool_array([[0, 0, 1], [1, 0, 0]]))

        ba1 = BitArray.from_bool_array([[1, 0, 0], [1, 1, 0]])
        ba2 = BitArray.from_bool_array([[1, 1, 0]])
        self.assertEqual(ba1 ^ ba2, BitArray.from_bool_array([[0, 1, 0], [0, 0, 0]]))

    def test_from_bool_array(self):
        """Test the from_bool_array static_constructor."""

        bit_array = BitArray.from_bool_array(
            [[[1, 0, 1, 0], [0, 0, 1, 1]], [[1, 0, 0, 0], [0, 0, 0, 1]]]
        )
        self.assertEqual(bit_array, BitArray(u_8([[[10], [3]], [[8], [1]]]), 4))

        bit_array = BitArray.from_bool_array(
            [[[1, 0, 1, 0], [0, 0, 1, 1]], [[1, 0, 0, 0], [0, 0, 0, 1]]], order="little"
        )
        self.assertEqual(bit_array, BitArray(u_8([[[5], [12]], [[1], [8]]]), 4))

        bit_array = BitArray.from_bool_array(
            [[0, 0, 1, 1, 1] + [0, 0, 0, 0, 0, 0, 1, 1] + [0, 0, 0, 0, 0, 0, 0, 1]]
        )
        self.assertEqual(bit_array, BitArray(u_8([[7, 3, 1]]), 21))

        bit_array = BitArray.from_bool_array(
            [[1, 0, 0, 0, 0, 0, 0, 0] + [1, 1, 0, 0, 0, 0, 0, 0] + [1, 1, 1, 0, 0]], order="little"
        )
        self.assertEqual(bit_array, BitArray(u_8([[7, 3, 1]]), 21))

        with self.assertRaisesRegex(ValueError, "unknown value for order"):
            BitArray.from_bool_array(
                [[[1, 0, 1, 0], [0, 0, 1, 1]], [[1, 0, 0, 0], [0, 0, 0, 1]]], order="litle"
            )

        with self.assertRaisesRegex(ValueError, "unknown value for order"):
            BitArray.from_bool_array(
                [[[1, 0, 1, 0], [0, 0, 1, 1]], [[1, 0, 0, 0], [0, 0, 0, 1]]], order="bg"
            )

    @ddt.data("counts", "int", "hex", "bit")
    def test_from_counts(self, counts_type):
        """Test the from_counts static constructor."""

        def convert(counts: Counts):
            if counts_type == "int":
                return counts.int_outcomes()
            if counts_type == "hex":
                return counts.hex_outcomes()
            if counts_type == "bit":
                return {bin(val): count for val, count in counts.int_outcomes().items()}
            return counts

        counts1 = convert(Counts({"0b101010": 2, "0b1": 3, "0x010203": 4}))
        counts2 = convert(Counts({1: 3, 2: 6}))
        counts3 = convert(Counts({0: 2}))

        bit_array = BitArray.from_counts(counts1)
        expected = BitArray(u_8([[0, 0, 42]] * 2 + [[0, 0, 1]] * 3 + [[1, 2, 3]] * 4), 17)
        self.assertEqual(bit_array, expected)

        bit_array = BitArray.from_counts(iter([counts1]))
        expected = BitArray(u_8([[[0, 0, 42]] * 2 + [[0, 0, 1]] * 3 + [[1, 2, 3]] * 4]), 17)
        self.assertEqual(bit_array, expected)

        bit_array = BitArray.from_counts(iter([counts1, counts2]))
        expected = [
            [[0, 0, 42]] * 2 + [[0, 0, 1]] * 3 + [[1, 2, 3]] * 4,
            [[0, 0, 1]] * 3 + [[0, 0, 2]] * 6,
        ]
        self.assertEqual(bit_array, BitArray(u_8(expected), 17))

        bit_array = BitArray.from_counts(counts3)
        expected = BitArray(u_8([[0], [0]]), 1)
        self.assertEqual(bit_array, expected)

    def test_from_samples_bitstring(self):
        """Test the from_samples static constructor."""
        bit_array = BitArray.from_samples(["110", "1", "1111111111"])
        self.assertEqual(bit_array, BitArray(u_8([[0, 6], [0, 1], [3, 255]]), 10))

        bit_array = BitArray.from_samples(["110", "1", "1111111111"], 20)
        self.assertEqual(bit_array, BitArray(u_8([[0, 0, 6], [0, 0, 1], [0, 3, 255]]), 20))

        bit_array = BitArray.from_samples(["000", "0"])
        self.assertEqual(bit_array, BitArray(u_8([[0], [0]]), 1))

    def test_from_samples_hex(self):
        """Test the from_samples static constructor."""
        bit_array = BitArray.from_samples(["0x01", "0x0a12", "0x0105"])
        self.assertEqual(bit_array, BitArray(u_8([[0, 1], [10, 18], [1, 5]]), 12))

        bit_array = BitArray.from_samples(["0x01", "0x0a12", "0x0105"], 20)
        self.assertEqual(bit_array, BitArray(u_8([[0, 0, 1], [0, 10, 18], [0, 1, 5]]), 20))

        bit_array = BitArray.from_samples(["0x0", "0x0"])
        self.assertEqual(bit_array, BitArray(u_8([[0], [0]]), 1))

    def test_from_samples_int(self):
        """Test the from_samples static constructor."""
        bit_array = BitArray.from_samples([1, 2578, 261])
        self.assertEqual(bit_array, BitArray(u_8([[0, 1], [10, 18], [1, 5]]), 12))

        bit_array = BitArray.from_samples([1, 2578, 261], 20)
        self.assertEqual(bit_array, BitArray(u_8([[0, 0, 1], [0, 10, 18], [0, 1, 5]]), 20))

        bit_array = BitArray.from_samples([0, 0, 0])
        self.assertEqual(bit_array, BitArray(u_8([[0], [0], [0]]), 1))

    def test_reshape(self):
        """Test the reshape method."""
        # this creates incrementing bitstrings from 0 to 360 * 32 - 1
        data = np.frombuffer(np.arange(360 * 32, dtype=np.uint64).tobytes(), dtype=np.uint8)
        data = data.reshape(-1, 32, 8)[..., 1::-1]
        ba = BitArray(data, 15)

        self.assertEqual(ba.reshape(120, 3).shape, (120, 3))
        self.assertEqual(ba.reshape(120, 3).num_shots, 32)
        self.assertEqual(ba.reshape(120, 3).num_bits, 15)
        self.assertTrue(
            np.array_equal(ba.reshape(60, 6).array[2, 3], data.reshape(60, 6, 32, 2)[2, 3])
        )

        self.assertEqual(ba.reshape(360 * 32).shape, ())
        self.assertEqual(ba.reshape(360 * 32).num_shots, 360 * 32)
        self.assertEqual(ba.reshape(360 * 32).num_bits, 15)

        self.assertEqual(ba.reshape(360 * 2, 16).shape, (720,))
        self.assertEqual(ba.reshape(360 * 2, 16).num_shots, 16)
        self.assertEqual(ba.reshape(360 * 2, 16).num_bits, 15)

    def test_transpose(self):
        """Test the transpose method."""
        # this creates incrementing bitstrings from 0 to 59
        data = np.frombuffer(np.arange(60, dtype=np.uint16).tobytes(), dtype=np.uint8)
        data = data.reshape(1, 2, 3, 10, 2)[..., ::-1]
        # Since the input dtype is uint16, bit array requires at least two u8.
        # Thus, 9 is the minimum number of qubits, i.e., 8 + 1.
        ba = BitArray(data, 9)
        self.assertEqual(ba.shape, (1, 2, 3))

        with self.subTest("default arg"):
            ba2 = ba.transpose()
            self.assertEqual(ba2.shape, (3, 2, 1))
            for i, j, k in product(range(1), range(2), range(3)):
                self.assertEqual(ba.get_counts((i, j, k)), ba2.get_counts((k, j, i)))

        with self.subTest("tuple 1"):
            ba2 = ba.transpose((2, 1, 0))
            self.assertEqual(ba2.shape, (3, 2, 1))
            for i, j, k in product(range(1), range(2), range(3)):
                self.assertEqual(ba.get_counts((i, j, k)), ba2.get_counts((k, j, i)))

        with self.subTest("tuple 2"):
            ba2 = ba.transpose((0, 1, 2))
            self.assertEqual(ba2.shape, (1, 2, 3))
            for i, j, k in product(range(1), range(2), range(3)):
                self.assertEqual(ba.get_counts((i, j, k)), ba2.get_counts((i, j, k)))

        with self.subTest("tuple 3"):
            ba2 = ba.transpose((0, 2, 1))
            self.assertEqual(ba2.shape, (1, 3, 2))
            for i, j, k in product(range(1), range(2), range(3)):
                self.assertEqual(ba.get_counts((i, j, k)), ba2.get_counts((i, k, j)))

        with self.subTest("tuple, negative indices"):
            ba2 = ba.transpose((0, -1, -2))
            self.assertEqual(ba2.shape, (1, 3, 2))
            for i, j, k in product(range(1), range(2), range(3)):
                self.assertEqual(ba.get_counts((i, j, k)), ba2.get_counts((i, k, j)))

        with self.subTest("ints"):
            ba2 = ba.transpose(2, 1, 0)
            self.assertEqual(ba2.shape, (3, 2, 1))
            for i, j, k in product(range(1), range(2), range(3)):
                self.assertEqual(ba.get_counts((i, j, k)), ba2.get_counts((k, j, i)))

        with self.subTest("errors"):
            with self.assertRaisesRegex(ValueError, "axes don't match bit array"):
                _ = ba.transpose((0, 1))
            with self.assertRaisesRegex(ValueError, "axes don't match bit array"):
                _ = ba.transpose((0, 1, 2, 3))
            with self.assertRaisesRegex(ValueError, "axis [0-9]+ is out of bounds for bit array"):
                _ = ba.transpose((0, 1, 4))
            with self.assertRaisesRegex(ValueError, "axis -[0-9]+ is out of bounds for bit array"):
                _ = ba.transpose((0, 1, -4))
            with self.assertRaisesRegex(ValueError, "repeated axis in transpose"):
                _ = ba.transpose((0, 1, 1))

    def test_concatenate(self):
        """Test the concatenate function."""
        # this creates incrementing bitstrings from 0 to 59
        data = np.frombuffer(np.arange(60, dtype=np.uint16).tobytes(), dtype=np.uint8)
        data = data.reshape(1, 2, 3, 10, 2)[..., ::-1]
        ba = BitArray(data, 9)
        self.assertEqual(ba.shape, (1, 2, 3))
        concatenate = BitArray.concatenate

        with self.subTest("2 arrays, default"):
            ba2 = concatenate([ba, ba])
            self.assertEqual(ba2.shape, (2, 2, 3))
            for j, k in product(range(2), range(3)):
                self.assertEqual(ba2.get_counts((0, j, k)), ba2.get_counts((1, j, k)))

        with self.subTest("2 arrays, axis"):
            ba2 = concatenate([ba, ba], axis=1)
            self.assertEqual(ba2.shape, (1, 4, 3))
            for j, k in product(range(2), range(3)):
                self.assertEqual(ba2.get_counts((0, j, k)), ba2.get_counts((0, j + 2, k)))

        with self.subTest("3 arrays"):
            ba2 = concatenate([ba, ba, ba])
            self.assertEqual(ba2.shape, (3, 2, 3))
            for j, k in product(range(2), range(3)):
                self.assertEqual(ba2.get_counts((0, j, k)), ba2.get_counts((1, j, k)))
                self.assertEqual(ba2.get_counts((1, j, k)), ba2.get_counts((2, j, k)))

        with self.subTest("errors"):
            with self.assertRaisesRegex(ValueError, "Need at least one bit array to concatenate"):
                _ = concatenate([])
            with self.assertRaisesRegex(ValueError, "axis -1 is out of bounds"):
                _ = concatenate([ba, ba], -1)
            with self.assertRaisesRegex(ValueError, "axis 100 is out of bounds"):
                _ = concatenate([ba, ba], 100)

            ba2 = BitArray(data, 10)
            with self.assertRaisesRegex(ValueError, "All bit arrays must have same number of bits"):
                _ = concatenate([ba, ba2])

            data2 = np.frombuffer(np.arange(30, dtype=np.uint16).tobytes(), dtype=np.uint8)
            data2 = data2.reshape(1, 2, 3, 5, 2)[..., ::-1]
            ba2 = BitArray(data2, 9)
            with self.assertRaisesRegex(
                ValueError, "All bit arrays must have same number of shots"
            ):
                _ = concatenate([ba, ba2])

            ba2 = ba.reshape(2, 3)
            with self.assertRaisesRegex(
                ValueError, "All bit arrays must have same number of dimensions"
            ):
                _ = concatenate([ba, ba2])

    def test_concatenate_shots(self):
        """Test the concatenate_shots function."""
        # this creates incrementing bitstrings from 0 to 59
        data = np.frombuffer(np.arange(60, dtype=np.uint16).tobytes(), dtype=np.uint8)
        data = data.reshape(1, 2, 3, 10, 2)[..., ::-1]
        ba = BitArray(data, 9)
        self.assertEqual(ba.shape, (1, 2, 3))
        concatenate_shots = BitArray.concatenate_shots

        with self.subTest("2 arrays"):
            ba2 = concatenate_shots([ba, ba])
            self.assertEqual(ba2.shape, (1, 2, 3))
            self.assertEqual(ba2.num_bits, 9)
            self.assertEqual(ba2.num_shots, 2 * ba.num_shots)
            for i, j, k in product(range(1), range(2), range(3)):
                expected = {key: val * 2 for key, val in ba.get_counts((i, j, k)).items()}
                counts2 = ba2.get_counts((i, j, k))
                self.assertEqual(counts2, expected)

        with self.subTest("3 arrays"):
            ba2 = concatenate_shots([ba, ba, ba])
            self.assertEqual(ba2.shape, (1, 2, 3))
            self.assertEqual(ba2.num_bits, 9)
            self.assertEqual(ba2.num_shots, 3 * ba.num_shots)
            for i, j, k in product(range(1), range(2), range(3)):
                expected = {key: val * 3 for key, val in ba.get_counts((i, j, k)).items()}
                counts2 = ba2.get_counts((i, j, k))
                self.assertEqual(counts2, expected)

        with self.subTest("errors"):
            with self.assertRaisesRegex(ValueError, "Need at least one bit array to stack"):
                _ = concatenate_shots([])

            ba2 = BitArray(data, 10)
            with self.assertRaisesRegex(ValueError, "All bit arrays must have same number of bits"):
                _ = concatenate_shots([ba, ba2])

            ba2 = ba.reshape(2, 3)
            with self.assertRaisesRegex(ValueError, "All bit arrays must have same shape"):
                _ = concatenate_shots([ba, ba2])

    def test_concatenate_bits(self):
        """Test the concatenate_bits function."""
        # this creates incrementing bitstrings from 0 to 59
        data = np.frombuffer(np.arange(60, dtype=np.uint16).tobytes(), dtype=np.uint8)
        data = data.reshape(1, 2, 3, 10, 2)[..., ::-1]
        ba = BitArray(data, 9)
        self.assertEqual(ba.shape, (1, 2, 3))
        concatenate_bits = BitArray.concatenate_bits

        with self.subTest("2 arrays"):
            ba_01 = ba.slice_bits([0, 1])
            ba2 = concatenate_bits([ba, ba_01])
            self.assertEqual(ba2.shape, (1, 2, 3))
            self.assertEqual(ba2.num_bits, 11)
            self.assertEqual(ba2.num_shots, ba.num_shots)
            for i, j, k in product(range(1), range(2), range(3)):
                bs = ba.get_bitstrings((i, j, k))
                bs_01 = ba_01.get_bitstrings((i, j, k))
                expected = [s1 + s2 for s1, s2 in zip(bs_01, bs)]
                bs2 = ba2.get_bitstrings((i, j, k))
                self.assertEqual(bs2, expected)

        with self.subTest("3 arrays"):
            ba_01 = ba.slice_bits([0, 1])
            ba2 = concatenate_bits([ba, ba_01, ba_01])
            self.assertEqual(ba2.shape, (1, 2, 3))
            self.assertEqual(ba2.num_bits, 13)
            self.assertEqual(ba2.num_shots, ba.num_shots)
            for i, j, k in product(range(1), range(2), range(3)):
                bs = ba.get_bitstrings((i, j, k))
                bs_01 = ba_01.get_bitstrings((i, j, k))
                expected = [s1 + s1 + s2 for s1, s2 in zip(bs_01, bs)]
                bs2 = ba2.get_bitstrings((i, j, k))
                self.assertEqual(bs2, expected)

        with self.subTest("errors"):
            with self.assertRaisesRegex(ValueError, "Need at least one bit array to stack"):
                _ = concatenate_bits([])

            data2 = np.frombuffer(np.arange(30, dtype=np.uint16).tobytes(), dtype=np.uint8)
            data2 = data2.reshape(1, 2, 3, 5, 2)[..., ::-1]
            ba2 = BitArray(data2, 9)
            with self.assertRaisesRegex(
                ValueError, "All bit arrays must have same number of shots"
            ):
                _ = concatenate_bits([ba, ba2])

            ba2 = ba.reshape(2, 3)
            with self.assertRaisesRegex(ValueError, "All bit arrays must have same shape"):
                _ = concatenate_bits([ba, ba2])

    def test_getitem(self):
        """Test the __getitem__ method."""
        # this creates incrementing bitstrings from 0 to 59
        data = np.frombuffer(np.arange(60, dtype=np.uint16).tobytes(), dtype=np.uint8)
        data = data.reshape(1, 2, 3, 10, 2)[..., ::-1]
        ba = BitArray(data, 9)
        self.assertEqual(ba.shape, (1, 2, 3))

        with self.subTest("all"):
            ba2 = ba[:]
            self.assertEqual(ba2.shape, (1, 2, 3))
            for i, j, k in product(range(1), range(2), range(3)):
                self.assertEqual(ba.get_counts((i, j, k)), ba2.get_counts((i, j, k)))

        with self.subTest("no slice"):
            ba2 = ba[0, 1, 2]
            self.assertEqual(ba2.shape, ())
            self.assertEqual(ba.get_counts((0, 1, 2)), ba2.get_counts())

        with self.subTest("slice"):
            ba2 = ba[0, :, 2]
            self.assertEqual(ba2.shape, (2,))
            for j in range(2):
                self.assertEqual(ba.get_counts((0, j, 2)), ba2.get_counts(j))

        with self.subTest("errors"):
            with self.assertRaisesRegex(IndexError, "index 2 is out of bounds"):
                _ = ba[0, 2, 2]
            with self.assertRaisesRegex(IndexError, "index -3 is out of bounds"):
                _ = ba[0, -3, 2]
            with self.assertRaisesRegex(
                IndexError, "BitArray cannot be sliced along the shots axis"
            ):
                _ = ba[0, 1, 2, 3]
            with self.assertRaisesRegex(
                IndexError, "BitArray cannot be sliced along the bits axis"
            ):
                _ = ba[0, 1, 2, 3, 4]

    def test_slice_bits(self):
        """Test the slice_bits method."""
        # this creates incrementing bitstrings from 0 to 59
        data = np.frombuffer(np.arange(60, dtype=np.uint16).tobytes(), dtype=np.uint8)
        data = data.reshape(1, 2, 3, 10, 2)[..., ::-1]
        ba = BitArray(data, 9)
        self.assertEqual(ba.shape, (1, 2, 3))

        with self.subTest("all"):
            ba2 = ba.slice_bits(range(ba.num_bits))
            self.assertEqual(ba2.shape, ba.shape)
            self.assertEqual(ba2.num_shots, ba.num_shots)
            self.assertEqual(ba2.num_bits, ba.num_bits)
            for i, j, k in product(range(1), range(2), range(3)):
                self.assertEqual(ba.get_counts((i, j, k)), ba2.get_counts((i, j, k)))

        with self.subTest("1 bit, int"):
            ba2 = ba.slice_bits(0)
            self.assertEqual(ba2.shape, ba.shape)
            self.assertEqual(ba2.num_shots, ba.num_shots)
            self.assertEqual(ba2.num_bits, 1)
            for i, j, k in product(range(1), range(2), range(3)):
                self.assertEqual(ba2.get_counts((i, j, k)), {"0": 5, "1": 5})

        with self.subTest("1 bit, list"):
            ba2 = ba.slice_bits([0])
            self.assertEqual(ba2.shape, ba.shape)
            self.assertEqual(ba2.num_shots, ba.num_shots)
            self.assertEqual(ba2.num_bits, 1)
            for i, j, k in product(range(1), range(2), range(3)):
                self.assertEqual(ba2.get_counts((i, j, k)), {"0": 5, "1": 5})

        with self.subTest("2 bits"):
            ba2 = ba.slice_bits([0, 1])
            self.assertEqual(ba2.shape, ba.shape)
            self.assertEqual(ba2.num_shots, ba.num_shots)
            self.assertEqual(ba2.num_bits, 2)
            even = {"00": 3, "01": 3, "10": 2, "11": 2}
            odd = {"10": 3, "11": 3, "00": 2, "01": 2}
            for count, (i, j, k) in enumerate(product(range(1), range(2), range(3))):
                expect = even if count % 2 == 0 else odd
                self.assertEqual(ba2.get_counts((i, j, k)), expect)

        with self.subTest("errors"):
            with self.assertRaisesRegex(IndexError, "index -1 is out of bounds"):
                _ = ba.slice_bits(-1)
            with self.assertRaisesRegex(IndexError, "index 9 is out of bounds"):
                _ = ba.slice_bits(9)

    def test_slice_shots(self):
        """Test the slice_shots method."""
        # this creates incrementing bitstrings from 0 to 59
        data = np.frombuffer(np.arange(60, dtype=np.uint16).tobytes(), dtype=np.uint8)
        data = data.reshape(1, 2, 3, 10, 2)[..., ::-1]
        ba = BitArray(data, 9)
        self.assertEqual(ba.shape, (1, 2, 3))

        with self.subTest("all"):
            ba2 = ba.slice_shots(range(ba.num_shots))
            self.assertEqual(ba2.shape, ba.shape)
            self.assertEqual(ba2.num_bits, ba.num_bits)
            self.assertEqual(ba2.num_shots, ba.num_shots)
            for i, j, k in product(range(1), range(2), range(3)):
                self.assertEqual(ba.get_counts((i, j, k)), ba2.get_counts((i, j, k)))

        with self.subTest("1 shot, int"):
            ba2 = ba.slice_shots(0)
            self.assertEqual(ba2.shape, ba.shape)
            self.assertEqual(ba2.num_bits, ba.num_bits)
            self.assertEqual(ba2.num_shots, 1)
            for i, j, k in product(range(1), range(2), range(3)):
                self.assertEqual(ba2.get_bitstrings((i, j, k)), [ba.get_bitstrings((i, j, k))[0]])

        with self.subTest("1 shot, list"):
            ba2 = ba.slice_shots([0])
            self.assertEqual(ba2.shape, ba.shape)
            self.assertEqual(ba2.num_bits, ba.num_bits)
            self.assertEqual(ba2.num_shots, 1)
            for i, j, k in product(range(1), range(2), range(3)):
                self.assertEqual(ba2.get_bitstrings((i, j, k)), [ba.get_bitstrings((i, j, k))[0]])

        with self.subTest("multiple shots"):
            indices = [1, 2, 3, 5, 8]
            ba2 = ba.slice_shots(indices)
            self.assertEqual(ba2.shape, ba.shape)
            self.assertEqual(ba2.num_bits, ba.num_bits)
            self.assertEqual(ba2.num_shots, len(indices))
            for i, j, k in product(range(1), range(2), range(3)):
                expected = [
                    bs for ind, bs in enumerate(ba.get_bitstrings((i, j, k))) if ind in indices
                ]
                self.assertEqual(ba2.get_bitstrings((i, j, k)), expected)

        with self.subTest("errors"):
            with self.assertRaisesRegex(IndexError, "index -1 is out of bounds"):
                _ = ba.slice_shots(-1)
            with self.assertRaisesRegex(IndexError, "index 10 is out of bounds"):
                _ = ba.slice_shots(10)

    def test_expectation_values(self):
        """Test the expectation_values method."""
        # this creates incrementing bitstrings from 0 to 59
        data = np.frombuffer(np.arange(60, dtype=np.uint16).tobytes(), dtype=np.uint8)
        data = data.reshape(1, 2, 3, 10, 2)[..., ::-1]
        ba = BitArray(data, 9)
        self.assertEqual(ba.shape, (1, 2, 3))
        op = "I" * 8 + "Z"
        op2 = "I" * 8 + "0"
        op3 = "I" * 8 + "1"
        pauli = Pauli(op)
        sp_op = SparsePauliOp(op)
        sp_op2 = SparsePauliOp.from_sparse_list([("Z", [6], 1)], num_qubits=9)

        with self.subTest("str"):
            expval = ba.expectation_values(op)
            # both 0 and 1 appear 5 times
            self.assertEqual(expval.shape, ba.shape)
            np.testing.assert_allclose(expval, np.zeros((ba.shape)))

            expval = ba.expectation_values(op2)
            self.assertEqual(expval.shape, ba.shape)
            np.testing.assert_allclose(expval, np.full((ba.shape), 0.5))

            expval = ba.expectation_values(op3)
            self.assertEqual(expval.shape, ba.shape)
            np.testing.assert_allclose(expval, np.full((ba.shape), 0.5))

            ba2 = ba.slice_bits(6)
            # 6th bit are all 0
            expval = ba2.expectation_values("Z")
            self.assertEqual(expval.shape, ba.shape)
            np.testing.assert_allclose(expval, np.ones(ba.shape))

            ba3 = ba.slice_bits(5)
            # 5th bit distributes as follows.
            # (0, 0, 0) {'0': 10}
            # (0, 0, 1) {'0': 10}
            # (0, 0, 2) {'0': 10}
            # (0, 1, 0) {'0': 2, '1': 8}
            # (0, 1, 1) {'1': 10}
            # (0, 1, 2) {'1': 10}
            expval = ba3.expectation_values("0")
            expected = np.array([[[1, 1, 1], [0.2, 0, 0]]])
            self.assertEqual(expval.shape, ba.shape)
            np.testing.assert_allclose(expval, expected)

        with self.subTest("Pauli"):
            expval = ba.expectation_values(pauli)
            self.assertEqual(expval.shape, ba.shape)
            np.testing.assert_allclose(expval, np.zeros((ba.shape)))

        with self.subTest("SparsePauliOp"):
            expval = ba.expectation_values(sp_op)
            self.assertEqual(expval.shape, ba.shape)
            np.testing.assert_allclose(expval, np.zeros((ba.shape)))

            expval = ba.expectation_values(sp_op2)
            # 6th bit are all 0
            self.assertEqual(expval.shape, ba.shape)
            np.testing.assert_allclose(expval, np.ones((ba.shape)))

        with self.subTest("ObservableArray"):
            obs = ["Z", "0", "1"]
            ba2 = ba.slice_bits(5)
            expval = ba2.expectation_values(obs)
            expected = np.array([[[1, 1, 0], [-0.6, 0, 1]]])
            self.assertEqual(expval.shape, ba.shape)
            np.testing.assert_allclose(expval, expected)

            ba4 = BitArray.from_counts([{0: 1}, {1: 1}]).reshape(2, 1)
            expval = ba4.expectation_values(obs)
            expected = np.array([[1, 1, 0], [-1, 0, 1]])
            self.assertEqual(expval.shape, (2, 3))
            np.testing.assert_allclose(expval, expected)

        with self.subTest("errors"):
            with self.assertRaisesRegex(ValueError, "shape mismatch"):
                _ = ba.expectation_values([op, op2])
            with self.assertRaisesRegex(ValueError, "One or more operators not same length"):
                _ = ba.expectation_values("Z")
            with self.assertRaisesRegex(ValueError, "is not diagonal"):
                _ = ba.expectation_values("X" * ba.num_bits)

    def test_postselection(self):
        """Test the postselection method."""

        flat_data = np.array(
            [
                [0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                [0, 0, 0, 0, 0, 1, 1, 1, 1, 1],
                [0, 1, 0, 1, 0, 1, 0, 1, 0, 1],
            ],
            dtype=bool,
        )

        shaped_data = np.array(
            [
                [
                    [
                        [0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                        [0, 0, 0, 0, 0, 1, 1, 1, 1, 1],
                        [0, 1, 0, 1, 0, 1, 0, 1, 0, 1],
                    ],
                    [
                        [1, 0, 1, 0, 1, 0, 1, 0, 1, 0],
                        [1, 1, 1, 1, 1, 0, 0, 0, 0, 0],
                        [1, 1, 1, 1, 1, 1, 1, 1, 1, 1],
                    ],
                ]
            ],
            dtype=bool,
        )

        for dataname, bool_array in zip(["flat", "shaped"], [flat_data, shaped_data]):

            bit_array = BitArray.from_bool_array(bool_array, order="little")
            # indices value of i <-> creg[i] <-> bool_array[..., i]

            num_bits = bool_array.shape[-1]
            bool_array = bool_array.reshape(-1, num_bits)

            test_cases = [
                ("basic", [0, 1], [0, 0]),
                ("multibyte", [0, 9], [0, 1]),
                ("repeated", [5, 5, 5], [0, 0, 0]),
                ("contradict", [5, 5, 5], [1, 0, 0]),
                ("unsorted", [5, 0, 9, 3], [1, 0, 1, 0]),
                ("negative", [-5, 1, -2, -10], [1, 0, 1, 0]),
                ("negcontradict", [4, -6], [1, 0]),
                ("trivial", [], []),
                ("bareindex", 6, 0),
            ]

            for name, indices, selection in test_cases:
                with self.subTest("_".join([dataname, name])):
                    postselected_bools = np.unpackbits(
                        bit_array.postselect(indices, selection).array[:, ::-1],
                        count=num_bits,
                        axis=-1,
                        bitorder="little",
                    ).astype(bool)
                    if isinstance(indices, int):
                        indices = (indices,)
                    if isinstance(selection, bool):
                        selection = (selection,)
                    answer = bool_array[np.all(bool_array[:, indices] == selection, axis=-1)]
                    if name in ["contradict", "negcontradict"]:
                        self.assertEqual(len(answer), 0)
                    else:
                        self.assertGreater(len(answer), 0)
                    np.testing.assert_equal(postselected_bools, answer)

            error_cases = [
                ("aboverange", [0, 6, 10], [True, True, False], IndexError),
                ("belowrange", [0, 6, -11], [True, True, False], IndexError),
                ("mismatch", [0, 1, 2], [False, False], ValueError),
            ]
            for name, indices, selection, error in error_cases:
                with self.subTest(dataname + "_" + name):
                    with self.assertRaises(error):
                        bit_array.postselect(indices, selection)



# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2024.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Test the LookaheadSwap pass"""

import unittest
from numpy import pi

from qiskit.dagcircuit import DAGCircuit
from qiskit.transpiler.passes import LookaheadSwap
from qiskit.transpiler import CouplingMap, Target
from qiskit.converters import circuit_to_dag
from qiskit.circuit.library import CXGate
from qiskit import ClassicalRegister, QuantumRegister, QuantumCircuit
from test import QiskitTestCase  # pylint: disable=wrong-import-order

from ..legacy_cmaps import MELBOURNE_CMAP


class TestLookaheadSwap(QiskitTestCase):
    """Tests the LookaheadSwap pass."""

    def test_lookahead_swap_doesnt_modify_mapped_circuit(self):
        """Test that lookahead swap is idempotent.

        It should not modify a circuit which is already compatible with the
        coupling map, and can be applied repeatedly without modifying the circuit.
        """

        qr = QuantumRegister(3, name="q")
        circuit = QuantumCircuit(qr)
        circuit.cx(qr[0], qr[2])
        circuit.cx(qr[0], qr[1])
        original_dag = circuit_to_dag(circuit)

        # Create coupling map which contains all two-qubit gates in the circuit.
        coupling_map = CouplingMap([[0, 1], [0, 2]])

        mapped_dag = LookaheadSwap(coupling_map).run(original_dag)

        self.assertEqual(original_dag, mapped_dag)

        remapped_dag = LookaheadSwap(coupling_map).run(mapped_dag)

        self.assertEqual(mapped_dag, remapped_dag)

    def test_lookahead_swap_should_add_a_single_swap(self):
        """Test that LookaheadSwap will insert a SWAP to match layout.

        For a single cx gate which is not available in the current layout, test
        that the mapper inserts a single swap to enable the gate.
        """

        qr = QuantumRegister(3, "q")
        circuit = QuantumCircuit(qr)
        circuit.cx(qr[0], qr[2])
        dag_circuit = circuit_to_dag(circuit)

        coupling_map = CouplingMap([[0, 1], [1, 2]])

        mapped_dag = LookaheadSwap(coupling_map).run(dag_circuit)

        self.assertEqual(
            mapped_dag.count_ops().get("swap", 0), dag_circuit.count_ops().get("swap", 0) + 1
        )

    def test_lookahead_swap_finds_minimal_swap_solution(self):
        """Of many valid SWAPs, test that LookaheadSwap finds the cheapest path.

        For a two CNOT circuit: cx q[0],q[2]; cx q[0],q[1]
        on the initial layout: qN -> qN
        (At least) two solutions exist:
        - SWAP q[0],[1], cx q[0],q[2], cx q[0],q[1]
        - SWAP q[1],[2], cx q[0],q[2], SWAP q[1],q[2], cx q[0],q[1]

        Verify that we find the first solution, as it requires fewer SWAPs.
        """

        qr = QuantumRegister(3, "q")
        circuit = QuantumCircuit(qr)
        circuit.cx(qr[0], qr[2])
        circuit.cx(qr[0], qr[1])

        dag_circuit = circuit_to_dag(circuit)

        coupling_map = CouplingMap([[0, 1], [1, 2]])

        mapped_dag = LookaheadSwap(coupling_map).run(dag_circuit)

        self.assertEqual(
            mapped_dag.count_ops().get("swap", 0), dag_circuit.count_ops().get("swap", 0) + 1
        )

    def test_lookahead_swap_maps_measurements(self):
        """Verify measurement nodes are updated to map correct cregs to re-mapped qregs.

        Create a circuit with measures on q0 and q2, following a swap between q0 and q2.
        Since that swap is not in the coupling, one of the two will be required to move.
        Verify that the mapped measure corresponds to one of the two possible layouts following
        the swap.

        """

        qr = QuantumRegister(3, "q")
        cr = ClassicalRegister(2)
        circuit = QuantumCircuit(qr, cr)

        circuit.cx(qr[0], qr[2])
        circuit.measure(qr[0], cr[0])
        circuit.measure(qr[2], cr[1])

        dag_circuit = circuit_to_dag(circuit)

        coupling_map = CouplingMap([[0, 1], [1, 2]])

        mapped_dag = LookaheadSwap(coupling_map).run(dag_circuit)

        mapped_measure_qargs = {op.qargs[0] for op in mapped_dag.named_nodes("measure")}

        self.assertIn(mapped_measure_qargs, [{qr[0], qr[1]}, {qr[1], qr[2]}])

    def test_lookahead_swap_maps_measurements_with_target(self):
        """Verify measurement nodes are updated to map correct cregs to re-mapped qregs.

        Create a circuit with measures on q0 and q2, following a swap between q0 and q2.
        Since that swap is not in the coupling, one of the two will be required to move.
        Verify that the mapped measure corresponds to one of the two possible layouts following
        the swap.

        """

        qr = QuantumRegister(3, "q")
        cr = ClassicalRegister(2)
        circuit = QuantumCircuit(qr, cr)

        circuit.cx(qr[0], qr[2])
        circuit.measure(qr[0], cr[0])
        circuit.measure(qr[2], cr[1])

        dag_circuit = circuit_to_dag(circuit)

        target = Target()
        target.add_instruction(CXGate(), {(0, 1): None, (1, 2): None})

        mapped_dag = LookaheadSwap(target).run(dag_circuit)

        mapped_measure_qargs = {op.qargs[0] for op in mapped_dag.named_nodes("measure")}

        self.assertIn(mapped_measure_qargs, [{qr[0], qr[1]}, {qr[1], qr[2]}])

    def test_lookahead_swap_maps_barriers(self):
        """Verify barrier nodes are updated to re-mapped qregs.

        Create a circuit with a barrier on q0 and q2, following a swap between q0 and q2.
        Since that swap is not in the coupling, one of the two will be required to move.
        Verify that the mapped barrier corresponds to one of the two possible layouts following
        the swap.

        """

        qr = QuantumRegister(3, "q")
        cr = ClassicalRegister(2)
        circuit = QuantumCircuit(qr, cr)

        circuit.cx(qr[0], qr[2])
        circuit.barrier(qr[0], qr[2])

        dag_circuit = circuit_to_dag(circuit)

        coupling_map = CouplingMap([[0, 1], [1, 2]])

        mapped_dag = LookaheadSwap(coupling_map).run(dag_circuit)

        mapped_barrier_qargs = [set(op.qargs) for op in mapped_dag.named_nodes("barrier")][0]

        self.assertIn(mapped_barrier_qargs, [{qr[0], qr[1]}, {qr[1], qr[2]}])

    def test_lookahead_swap_higher_depth_width_is_better(self):
        """Test that lookahead swap finds better circuit with increasing search space.

        Increasing the tree width and depth is expected to yield a better (or same) quality
        circuit, in the form of fewer SWAPs.
        """
        # q_0: ──■───────────────────■───────────────────────────────────────────────»
        #      ┌─┴─┐                 │                 ┌───┐                         »
        # q_1: ┤ X ├──■──────────────┼─────────────────┤ X ├─────────────────────────»
        #      └───┘┌─┴─┐            │                 └─┬─┘┌───┐          ┌───┐     »
        # q_2: ─────┤ X ├──■─────────┼───────────────────┼──┤ X ├──────────┤ X ├──■──»
        #           └───┘┌─┴─┐     ┌─┴─┐                 │  └─┬─┘     ┌───┐└─┬─┘  │  »
        # q_3: ──────────┤ X ├──■──┤ X ├─────────────────┼────┼────■──┤ X ├──┼────┼──»
        #                └───┘┌─┴─┐└───┘          ┌───┐  │    │    │  └─┬─┘  │    │  »
        # q_4: ───────────────┤ X ├──■────────────┤ X ├──┼────■────┼────┼────┼────┼──»
        #                     └───┘┌─┴─┐          └─┬─┘  │         │    │    │    │  »
        # q_5: ────────────────────┤ X ├──■─────────┼────┼─────────┼────■────┼────┼──»
        #                          └───┘┌─┴─┐       │    │         │         │    │  »
        # q_6: ─────────────────────────┤ X ├──■────■────┼─────────┼─────────■────┼──»
        #                               └───┘┌─┴─┐       │       ┌─┴─┐          ┌─┴─┐»
        # q_7: ──────────────────────────────┤ X ├───────■───────┤ X ├──────────┤ X ├»
        #                                    └───┘               └───┘          └───┘»
        # «q_0: ──■───────
        # «       │
        # «q_1: ──┼───────
        # «       │
        # «q_2: ──┼───────
        # «       │
        # «q_3: ──┼───────
        # «       │
        # «q_4: ──┼───────
        # «       │
        # «q_5: ──┼────■──
        # «     ┌─┴─┐  │
        # «q_6: ┤ X ├──┼──
        # «     └───┘┌─┴─┐
        # «q_7: ─────┤ X ├
        # «          └───┘
        qr = QuantumRegister(8, name="q")
        circuit = QuantumCircuit(qr)
        circuit.cx(qr[0], qr[1])
        circuit.cx(qr[1], qr[2])
        circuit.cx(qr[2], qr[3])
        circuit.cx(qr[3], qr[4])
        circuit.cx(qr[4], qr[5])
        circuit.cx(qr[5], qr[6])
        circuit.cx(qr[6], qr[7])
        circuit.cx(qr[0], qr[3])
        circuit.cx(qr[6], qr[4])
        circuit.cx(qr[7], qr[1])
        circuit.cx(qr[4], qr[2])
        circuit.cx(qr[3], qr[7])
        circuit.cx(qr[5], qr[3])
        circuit.cx(qr[6], qr[2])
        circuit.cx(qr[2], qr[7])
        circuit.cx(qr[0], qr[6])
        circuit.cx(qr[5], qr[7])
        original_dag = circuit_to_dag(circuit)

        # Create a ring of 8 connected qubits
        coupling_map = CouplingMap.from_grid(num_rows=2, num_columns=4)

        mapped_dag_1 = LookaheadSwap(coupling_map, search_depth=3, search_width=3).run(original_dag)
        mapped_dag_2 = LookaheadSwap(coupling_map, search_depth=5, search_width=5).run(original_dag)

        num_swaps_1 = mapped_dag_1.count_ops().get("swap", 0)
        num_swaps_2 = mapped_dag_2.count_ops().get("swap", 0)

        self.assertLessEqual(num_swaps_2, num_swaps_1)

    def test_lookahead_swap_hang_in_min_case(self):
        """Verify LookaheadSwap does not stall in minimal case."""
        # ref: https://github.com/Qiskit/qiskit-terra/issues/2171

        qr = QuantumRegister(14, "q")
        qc = QuantumCircuit(qr)
        qc.cx(qr[0], qr[13])
        qc.cx(qr[1], qr[13])
        qc.cx(qr[1], qr[0])
        qc.cx(qr[13], qr[1])
        dag = circuit_to_dag(qc)

        cmap = CouplingMap(MELBOURNE_CMAP)
        out = LookaheadSwap(cmap, search_depth=4, search_width=4).run(dag)

        self.assertIsInstance(out, DAGCircuit)

    def test_lookahead_swap_hang_full_case(self):
        """Verify LookaheadSwap does not stall in reported case."""
        # ref: https://github.com/Qiskit/qiskit-terra/issues/2171

        qr = QuantumRegister(14, "q")
        qc = QuantumCircuit(qr)
        qc.cx(qr[0], qr[13])
        qc.cx(qr[1], qr[13])
        qc.cx(qr[1], qr[0])
        qc.cx(qr[13], qr[1])
        qc.cx(qr[6], qr[7])
        qc.cx(qr[8], qr[7])
        qc.cx(qr[8], qr[6])
        qc.cx(qr[7], qr[8])
        qc.cx(qr[0], qr[13])
        qc.cx(qr[1], qr[0])
        qc.cx(qr[13], qr[1])
        qc.cx(qr[0], qr[1])
        dag = circuit_to_dag(qc)

        cmap = CouplingMap(MELBOURNE_CMAP)

        out = LookaheadSwap(cmap, search_depth=4, search_width=4).run(dag)

        self.assertIsInstance(out, DAGCircuit)

    def test_global_phase_preservation(self):
        """Test that LookaheadSwap preserves global phase"""

        qr = QuantumRegister(3, "q")
        circuit = QuantumCircuit(qr)
        circuit.global_phase = pi / 3
        circuit.cx(qr[0], qr[2])
        dag_circuit = circuit_to_dag(circuit)

        coupling_map = CouplingMap([[0, 1], [1, 2]])

        mapped_dag = LookaheadSwap(coupling_map).run(dag_circuit)

        self.assertEqual(mapped_dag.global_phase, circuit.global_phase)
        self.assertEqual(
            mapped_dag.count_ops().get("swap", 0), dag_circuit.count_ops().get("swap", 0) + 1
        )


if __name__ == "__main__":
    unittest.main()



# This code is part of Qiskit.
#
# (C) Copyright IBM 2020.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""
Drawing objects for timeline drawer.

Drawing objects play two important roles:
    - Allowing unittests of visualization module. Usually it is hard for image files to be tested.
    - Removing program parser from each plotter interface. We can easily add new plotter.

This module is based on the structure of matplotlib as it is the primary plotter
of the timeline drawer. However this interface is agnostic to the actual plotter.

Design concept
~~~~~~~~~~~~~~
When we think about dynamically updating drawings, it will be most efficient to
update only the changed properties of drawings rather than regenerating entirely from scratch.
Thus the core :py:class:`~qiskit.visualization.timeline.core.DrawerCanvas` generates
all possible drawings in the beginning and then the canvas instance manages
visibility of each drawing according to the end-user request.

Data key
~~~~~~~~
In the abstract class ``ElementaryData`` common attributes to represent a drawing are
specified. In addition, drawings have the `data_key` property that returns an
unique hash of the object for comparison.
This key is generated from a data type, the location of the drawing in the canvas,
and associated qubit or classical bit objects.
See py:mod:`qiskit.visualization.timeline.types` for detail on the data type.
If a data key cannot distinguish two independent objects, you need to add a new data type.
The data key may be used in the plotter interface to identify the object.

Drawing objects
~~~~~~~~~~~~~~~
To support not only `matplotlib` but also multiple plotters, those drawings should be
universal and designed without strong dependency on modules in `matplotlib`.
This means drawings that represent primitive geometries are preferred.
It should be noted that there will be no unittest for each plotter API, which takes
drawings and outputs image data, we should avoid adding a complicated geometry
that has a context of the scheduled circuit program.

For example, a two qubit scheduled gate may be drawn by two rectangles that represent
time occupation of two quantum registers during the gate along with a line connecting
these rectangles to identify the pair. This shape can be represented with
two box-type objects with one line-type object instead of defining a new object dedicated
to the two qubit gate. As many plotters don't support an API that visualizes such
a linked-box shape, if we introduce such complex drawings and write a
custom wrapper function on top of the existing API,
it could be difficult to prevent bugs with the CI tools due to lack of
the effective unittest for image data.

Link between gates
~~~~~~~~~~~~~~~~~~
The ``GateLinkData`` is the special subclass of drawing that represents
a link between bits. Usually objects are associated to the specific bit,
but ``GateLinkData`` can be associated with multiple bits to illustrate relationship
between quantum or classical bits during a gate operation.
"""

from abc import ABC
from enum import Enum
from typing import Optional, Dict, Any, List, Union

import numpy as np

from qiskit import circuit
from qiskit.visualization.timeline import types
from qiskit.visualization.exceptions import VisualizationError


class ElementaryData(ABC):
    """Base class of the scheduled circuit visualization object.

    Note that drawings are mutable.
    """

    __hash__ = None

    def __init__(
        self,
        data_type: Union[str, Enum],
        xvals: Union[np.ndarray, List[types.Coordinate]],
        yvals: Union[np.ndarray, List[types.Coordinate]],
        bits: Optional[Union[types.Bits, List[types.Bits]]] = None,
        meta: Optional[Dict[str, Any]] = None,
        styles: Optional[Dict[str, Any]] = None,
    ):
        """Create new drawing.

        Args:
            data_type: String representation of this drawing.
            xvals: Series of horizontal coordinate that the object is drawn.
            yvals: Series of vertical coordinate that the object is drawn.
            bits: Qubit or Clbit object bound to this drawing.
            meta: Meta data dictionary of the object.
            styles: Style keyword args of the object. This conforms to `matplotlib`.
        """
        if bits and isinstance(bits, (circuit.Qubit, circuit.Clbit)):
            bits = [bits]

        if isinstance(data_type, Enum):
            data_type = data_type.value

        self.data_type = str(data_type)
        self.xvals = xvals
        self.yvals = yvals
        self.bits = bits
        self.meta = meta
        self.styles = styles

    @property
    def data_key(self):
        """Return unique hash of this object."""
        return str(
            hash(
                (
                    self.__class__.__name__,
                    self.data_type,
                    tuple(self.bits),
                    tuple(self.xvals),
                    tuple(self.yvals),
                )
            )
        )

    def __repr__(self):
        return f"{self.__class__.__name__}(type={self.data_type}, key={self.data_key})"

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.data_key == other.data_key


class LineData(ElementaryData):
    """Drawing object that represents line shape."""

    def __init__(
        self,
        data_type: Union[str, Enum],
        xvals: Union[np.ndarray, List[types.Coordinate]],
        yvals: Union[np.ndarray, List[types.Coordinate]],
        bit: types.Bits,
        meta: Dict[str, Any] = None,
        styles: Dict[str, Any] = None,
    ):
        """Create new line.

        Args:
            data_type: String representation of this drawing.
            xvals: Series of horizontal coordinate that the object is drawn.
            yvals: Series of vertical coordinate that the object is drawn.
            bit: Bit associated to this object.
            meta: Meta data dictionary of the object.
            styles: Style keyword args of the object. This conforms to `matplotlib`.
        """
        super().__init__(
            data_type=data_type, xvals=xvals, yvals=yvals, bits=bit, meta=meta, styles=styles
        )


class BoxData(ElementaryData):
    """Drawing object that represents box shape."""

    def __init__(
        self,
        data_type: Union[str, Enum],
        xvals: Union[np.ndarray, List[types.Coordinate]],
        yvals: Union[np.ndarray, List[types.Coordinate]],
        bit: types.Bits,
        meta: Dict[str, Any] = None,
        styles: Dict[str, Any] = None,
    ):
        """Create new box.

        Args:
            data_type: String representation of this drawing.
            xvals: Left and right coordinate that the object is drawn.
            yvals: Top and bottom coordinate that the object is drawn.
            bit: Bit associated to this object.
            meta: Meta data dictionary of the object.
            styles: Style keyword args of the object. This conforms to `matplotlib`.

        Raises:
            VisualizationError: When number of data points are not equals to 2.
        """
        if len(xvals) != 2 or len(yvals) != 2:
            raise VisualizationError("Length of data points are not equals to 2.")

        super().__init__(
            data_type=data_type, xvals=xvals, yvals=yvals, bits=bit, meta=meta, styles=styles
        )


class TextData(ElementaryData):
    """Drawing object that represents a text on canvas."""

    def __init__(
        self,
        data_type: Union[str, Enum],
        xval: types.Coordinate,
        yval: types.Coordinate,
        bit: types.Bits,
        text: str,
        latex: Optional[str] = None,
        meta: Dict[str, Any] = None,
        styles: Dict[str, Any] = None,
    ):
        """Create new text.

        Args:
            data_type: String representation of this drawing.
            xval: Horizontal coordinate that the object is drawn.
            yval: Vertical coordinate that the object is drawn.
            bit: Bit associated to this object.
            text: A string to draw on the canvas.
            latex: If set this string is used instead of `text`.
            meta: Meta data dictionary of the object.
            styles: Style keyword args of the object. This conforms to `matplotlib`.
        """
        self.text = text
        self.latex = latex

        super().__init__(
            data_type=data_type, xvals=[xval], yvals=[yval], bits=bit, meta=meta, styles=styles
        )


class GateLinkData(ElementaryData):
    """A special drawing data type that represents bit link of multi-bit gates.

    Note this object takes multiple bits and dedicates them to the bit link.
    This may appear as a line on the canvas.
    """

    def __init__(
        self, xval: types.Coordinate, bits: List[types.Bits], styles: Dict[str, Any] = None
    ):
        """Create new bit link.

        Args:
            xval: Horizontal coordinate that the object is drawn.
            bits: Bit associated to this object.
            styles: Style keyword args of the object. This conforms to `matplotlib`.
        """
        super().__init__(
            data_type=types.LineType.GATE_LINK,
            xvals=[xval],
            yvals=[0],
            bits=bits,
            meta=None,
            styles=styles,
        )



# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2020.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Multiple-Control, Multiple-Target Gate."""

from __future__ import annotations

import warnings
from collections.abc import Callable

from qiskit import circuit
from qiskit.circuit import ControlledGate, Gate, QuantumCircuit
from qiskit.circuit._utils import _ctrl_state_to_int
from qiskit.utils.deprecation import deprecate_func
from ..standard_gates import get_standard_gate_name_mapping


class MCMT(QuantumCircuit):
    """The multi-controlled multi-target gate, for an arbitrary singly controlled target gate.

    For example, the H gate controlled on 3 qubits and acting on 2 target qubit is represented as:

    .. code-block:: text

        ───■────
           │
        ───■────
           │
        ───■────
        ┌──┴───┐
        ┤0     ├
        │  2-H │
        ┤1     ├
        └──────┘

    This default implementations requires no ancilla qubits, by broadcasting the target gate
    to the number of target qubits and using Qiskit's generic control routine to control the
    broadcasted target on the control qubits. If ancilla qubits are available, a more efficient
    variant using the so-called V-chain decomposition can be used. This is implemented in
    :class:`~qiskit.circuit.library.MCMTVChain`.
    """

    @deprecate_func(since="1.4", additional_msg="Use MCMTGate instead.")
    def __init__(
        self,
        gate: Gate | Callable[[QuantumCircuit, circuit.Qubit, circuit.Qubit], circuit.Instruction],
        num_ctrl_qubits: int,
        num_target_qubits: int,
    ) -> None:
        """Create a new multi-control multi-target gate.

        Args:
            gate: The gate to be applied controlled on the control qubits and applied to the target
                qubits. Can be either a Gate or a circuit method.
                If it is a callable, it will be casted to a Gate.
            num_ctrl_qubits: The number of control qubits.
            num_target_qubits: The number of target qubits.

        Raises:
            AttributeError: If the gate cannot be casted to a controlled gate.
            AttributeError: If the number of controls or targets is 0.
        """
        if num_ctrl_qubits == 0 or num_target_qubits == 0:
            raise AttributeError("Need at least one control and one target qubit.")

        if callable(gate):
            warnings.warn(
                "Passing a callable to MCMT is pending deprecation since Qiskit 1.3. Pass a "
                "gate instance or the gate name instead, e.g. pass 'h' instead of QuantumCircuit.h.",
                category=DeprecationWarning,
                stacklevel=2,
            )
            gate = gate.__name__
        elif isinstance(gate, QuantumCircuit):
            warnings.warn(
                "Passing a QuantumCircuit is pending deprecation since Qiskit 1.3. Pass a gate "
                "or turn the circuit into a gate using the ``to_gate`` method, instead.",
                category=DeprecationWarning,
                stacklevel=2,
            )
            gate = gate.to_gate()

        self.gate = MCMTGate._identify_base_gate(gate)
        self.num_ctrl_qubits = num_ctrl_qubits
        self.num_target_qubits = num_target_qubits

        # initialize the circuit object
        num_qubits = num_ctrl_qubits + num_target_qubits + self.num_ancilla_qubits
        super().__init__(num_qubits, name="mcmt")
        self._build()

    def _build(self):
        gate = MCMTGate(self.gate, self.num_ctrl_qubits, self.num_target_qubits)
        self.append(gate, self.qubits)

    @property
    def num_ancilla_qubits(self):
        """Return the number of ancillas."""
        return 0

    def control(self, num_ctrl_qubits=1, label=None, ctrl_state=None, annotated=False):
        """Return the controlled version of the MCMT circuit."""
        if not annotated and ctrl_state is None:
            gate = MCMT(self.gate, self.num_ctrl_qubits + num_ctrl_qubits, self.num_target_qubits)
        else:
            gate = super().control(num_ctrl_qubits, label, ctrl_state, annotated=annotated)
        return gate

    def inverse(self, annotated: bool = False):
        """Return the inverse MCMT circuit, which is itself."""
        return MCMT(self.gate, self.num_ctrl_qubits, self.num_target_qubits)


class MCMTVChain(MCMT):
    """The MCMT implementation using the CCX V-chain.

    This implementation requires ancillas but is decomposed into a much shallower circuit
    than the default implementation in :class:`~qiskit.circuit.library.MCMT`.

    **Expanded Circuit:**

    .. plot::
       :alt: Diagram illustrating the previously described circuit.

       from qiskit.circuit.library import MCMTVChain, ZGate
       from qiskit.visualization.library import _generate_circuit_library_visualization
       circuit = MCMTVChain(ZGate(), 2, 2)
       _generate_circuit_library_visualization(circuit.decompose())

    **Examples:**

        >>> from qiskit.circuit.library import HGate
        >>> MCMTVChain(HGate(), 3, 2).draw()

        q_0: ──■────────────────────────■──
               │                        │
        q_1: ──■────────────────────────■──
               │                        │
        q_2: ──┼────■──────────────■────┼──
               │    │  ┌───┐       │    │
        q_3: ──┼────┼──┤ H ├───────┼────┼──
               │    │  └─┬─┘┌───┐  │    │
        q_4: ──┼────┼────┼──┤ H ├──┼────┼──
             ┌─┴─┐  │    │  └─┬─┘  │  ┌─┴─┐
        q_5: ┤ X ├──■────┼────┼────■──┤ X ├
             └───┘┌─┴─┐  │    │  ┌─┴─┐└───┘
        q_6: ─────┤ X ├──■────■──┤ X ├─────
                  └───┘          └───┘
    """

    @deprecate_func(
        since="1.4",
        additional_msg="Use MCMTGate with the V-chain synthesis plugin instead.",
    )
    def __init__(
        self,
        gate: Gate | Callable[[QuantumCircuit, circuit.Qubit, circuit.Qubit], circuit.Instruction],
        num_ctrl_qubits: int,
        num_target_qubits: int,
    ) -> None:
        super().__init__(gate, num_ctrl_qubits, num_target_qubits)

    def _build(self):
        # pylint: disable=cyclic-import
        from qiskit.synthesis.multi_controlled import synth_mcmt_vchain

        synthesized = synth_mcmt_vchain(self.gate, self.num_ctrl_qubits, self.num_target_qubits)
        self.compose(synthesized, inplace=True, copy=False)

    @property
    def num_ancilla_qubits(self):
        """Return the number of ancilla qubits required."""
        return max(0, self.num_ctrl_qubits - 1)

    def inverse(self, annotated: bool = False):
        return MCMTVChain(self.gate, self.num_ctrl_qubits, self.num_target_qubits)


class MCMTGate(ControlledGate):
    """The multi-controlled multi-target gate, for an arbitrary singly controlled target gate.

    For example, the H gate controlled on 3 qubits and acting on 2 target qubit is represented as:

    .. parsed-literal::

        ───■────
           │
        ───■────
           │
        ───■────
        ┌──┴───┐
        ┤0     ├
        │  2-H │
        ┤1     ├
        └──────┘

    Depending on the number of available auxiliary qubits, this operation can be synthesized
    using different methods. For example, if :math:`n - 1` clean auxiliary qubits are available
    (where :math:`n` is the number of control qubits), a V-chain decomposition can be used whose
    depth is linear in :math:`n`. See also :func:`.synth_mcmt_chain`.
    """

    def __init__(
        self,
        gate: Gate,
        num_ctrl_qubits: int,
        num_target_qubits: int,
        ctrl_state: int | str | None = None,
        label: str | None = None,
    ) -> None:
        """
        Args:
            gate: The base gate to apply on multiple target qubits, controlled by other qubits.
                This must be a single-qubit gate or a controlled single-qubit gate.
            num_ctrl_qubits: The number of control qubits.
            num_target_qubits: The number of target qubits.
            ctrl_state: The control state of the control qubits. Defaults to all closed controls.
            label: The gate label.
        """
        if num_target_qubits < 1:
            raise ValueError("Need at least one target qubit.")

        if num_ctrl_qubits < 1:
            raise ValueError("Need at least one control qubit.")

        self.num_target_qubits = num_target_qubits

        base_gate = self._identify_base_gate(gate)
        num_qubits = num_ctrl_qubits + num_target_qubits

        if label is None:
            label = f"{num_target_qubits}-{gate.name.capitalize()}"

        super().__init__(
            "mcmt",
            base_gate=base_gate,
            num_qubits=num_qubits,
            params=gate.params,
            num_ctrl_qubits=num_ctrl_qubits,
            ctrl_state=ctrl_state,
            label=label,
        )

    def _define(self):
        """Default definition relying on gate.control. Control state is handled by superclass."""
        # pylint: disable=cyclic-import
        from qiskit.transpiler.passes.synthesis.hls_plugins import MCMTSynthesisDefault

        self.definition = MCMTSynthesisDefault().run(self)

    @staticmethod
    def _identify_base_gate(gate):
        """Get the control base gate. Note this must be a single qubit gate."""

        # try getting the standard name from the string
        if isinstance(gate, str):
            standard_gates = get_standard_gate_name_mapping()
            if gate in standard_gates:
                gate = standard_gates[gate]
            else:
                raise AttributeError(
                    f"Unknown gate {gate}. Available: {list(get_standard_gate_name_mapping.keys())}"
                )

        # extract the base gate
        if isinstance(gate, ControlledGate):
            warnings.warn(
                "Passing a controlled gate to MCMT is pending deprecation since Qiskit 1.3. Pass a "
                "single-qubit gate instance or the gate name instead, e.g. pass 'h' instead of 'ch'.",
                category=DeprecationWarning,
                stacklevel=2,
            )
            base_gate = gate.base_gate
        elif isinstance(gate, Gate):
            base_gate = gate
        else:
            raise TypeError(f"Invalid gate type {type(gate)}.")

        if base_gate.num_qubits != 1:
            raise ValueError(
                f"MCMTGate requires a base gate with a single qubit, but got {base_gate.num_qubits}."
            )

        return base_gate

    def control(self, num_ctrl_qubits=1, label=None, ctrl_state=None, annotated=False):
        """Return the controlled version of the MCMT circuit."""
        if not annotated:
            ctrl_state = _ctrl_state_to_int(ctrl_state, num_ctrl_qubits)
            new_ctrl_state = (self.ctrl_state << num_ctrl_qubits) | ctrl_state

            gate = MCMTGate(
                self.base_gate,
                self.num_ctrl_qubits + num_ctrl_qubits,
                self.num_target_qubits,
                ctrl_state=new_ctrl_state,
            )
        else:
            gate = super().control(num_ctrl_qubits, label, ctrl_state, annotated=annotated)

        return gate

    def inverse(self, annotated: bool = False):
        """Return the inverse MCMT circuit."""
        return MCMTGate(
            self.base_gate.inverse(), self.num_ctrl_qubits, self.num_target_qubits, self.ctrl_state
        )



# This code is part of Qiskit.
#
# (C) Copyright IBM 2022, 2023.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
"""
Utility functions for primitives
"""
from __future__ import annotations

from collections.abc import Iterable

import numpy as np

from qiskit.circuit import Instruction, QuantumCircuit
from qiskit.circuit.bit import Bit
from qiskit.circuit.library.data_preparation import Initialize
from qiskit.exceptions import QiskitError
from qiskit.quantum_info import PauliList, SparsePauliOp, Statevector
from qiskit.quantum_info.operators.base_operator import BaseOperator
from qiskit.quantum_info.operators.symplectic.base_pauli import BasePauli
from qiskit.utils.deprecation import deprecate_func


@deprecate_func(
    since="1.2",
    additional_msg="To initialize a circuit from a ``Statevector`` instance, "
    + "use ``QuantumCircuit.initialize`` instead.",
)
def init_circuit(state: QuantumCircuit | Statevector) -> QuantumCircuit:
    """Initialize state by converting the input to a quantum circuit.

    Args:
        state: The state as quantum circuit or statevector.

    Returns:
        The state as quantum circuit.
    """
    if isinstance(state, QuantumCircuit):
        return state
    if not isinstance(state, Statevector):
        state = Statevector(state)
    qc = QuantumCircuit(state.num_qubits)
    qc.append(Initialize(state), qargs=range(state.num_qubits))
    return qc


@deprecate_func(
    since="1.2",
    additional_msg="Use the constructor of ``SparsePauliOp`` instead.",
)
def init_observable(observable: BaseOperator | str) -> SparsePauliOp:
    """Initialize observable by converting the input to a :class:`~qiskit.quantum_info.SparsePauliOp`.

    Args:
        observable: The observable.

    Returns:
        The observable as :class:`~qiskit.quantum_info.SparsePauliOp`.

    Raises:
        QiskitError: when observable type cannot be converted to SparsePauliOp.
    """

    if isinstance(observable, SparsePauliOp):
        return observable
    elif isinstance(observable, BaseOperator) and not isinstance(observable, BasePauli):
        raise QiskitError(f"observable type not supported: {type(observable)}")
    else:
        if isinstance(observable, PauliList):
            raise QiskitError(f"observable type not supported: {type(observable)}")
        return SparsePauliOp(observable)


@deprecate_func(
    since="1.2",
    additional_msg="Use ``QuantumCircuit.layout`` and ``SparsePauliOp.apply_layout`` "
    + "to adjust an operator for a layout. Otherwise, use ``mthree.utils.final_measurement_mapping``. "
    + "See <https://qiskit.github.io/qiskit-addon-mthree/apidocs/utils> for details.",
)
def final_measurement_mapping(circuit: QuantumCircuit) -> dict[int, int]:
    """Return the final measurement mapping for the circuit.

    Dict keys label measured qubits, whereas the values indicate the
    classical bit onto which that qubits measurement result is stored.

    Parameters:
        circuit: Input quantum circuit.

    Returns:
        Mapping of qubits to classical bits for final measurements.
    """
    active_qubits = list(range(circuit.num_qubits))
    active_cbits = list(range(circuit.num_clbits))

    # Find final measurements starting in back
    mapping = {}
    for item in circuit._data[::-1]:
        if item.operation.name == "measure":
            cbit = circuit.find_bit(item.clbits[0]).index
            qbit = circuit.find_bit(item.qubits[0]).index
            if cbit in active_cbits and qbit in active_qubits:
                mapping[qbit] = cbit
                active_cbits.remove(cbit)
                active_qubits.remove(qbit)
        elif item.operation.name not in ["barrier", "delay"]:
            for qq in item.qubits:
                _temp_qubit = circuit.find_bit(qq).index
                if _temp_qubit in active_qubits:
                    active_qubits.remove(_temp_qubit)

        if not active_cbits or not active_qubits:
            break

    # Sort so that classical bits are in numeric order low->high.
    mapping = dict(sorted(mapping.items(), key=lambda item: item[1]))
    return mapping


def _bits_key(bits: tuple[Bit, ...], circuit: QuantumCircuit) -> tuple:
    return tuple(
        (
            circuit.find_bit(bit).index,
            tuple((reg[0].size, reg[0].name, reg[1]) for reg in circuit.find_bit(bit).registers),
        )
        for bit in bits
    )


def _format_params(param):
    if isinstance(param, np.ndarray):
        return param.data.tobytes()
    elif isinstance(param, QuantumCircuit):
        return _circuit_key(param)
    elif isinstance(param, Iterable):
        return tuple(param)
    return param


def _circuit_key(circuit: QuantumCircuit, functional: bool = True) -> tuple:
    """Private key function for QuantumCircuit.

    This is the workaround until :meth:`QuantumCircuit.__hash__` will be introduced.
    If key collision is found, please add elements to avoid it.

    Args:
        circuit: Input quantum circuit.
        functional: If True, the returned key only includes functional data (i.e. execution related).

    Returns:
        Composite key for circuit.
    """
    functional_key: tuple = (
        circuit.num_qubits,
        circuit.num_clbits,
        circuit.num_parameters,
        tuple(  # circuit.data
            (
                _bits_key(data.qubits, circuit),  # qubits
                _bits_key(data.clbits, circuit),  # clbits
                data.operation.name,  # operation.name
                tuple(_format_params(param) for param in data.operation.params),  # operation.params
            )
            for data in circuit.data
        ),
        None if circuit._op_start_times is None else tuple(circuit._op_start_times),
    )
    if functional:
        return functional_key
    return (
        circuit.name,
        *functional_key,
    )


def _observable_key(observable: SparsePauliOp) -> tuple:
    """Private key function for SparsePauliOp.
    Args:
        observable: Input operator.

    Returns:
        Key for observables.
    """
    return (
        observable.paulis.z.tobytes(),
        observable.paulis.x.tobytes(),
        observable.paulis.phase.tobytes(),
        observable.coeffs.tobytes(),
    )


def bound_circuit_to_instruction(circuit: QuantumCircuit) -> Instruction:
    """Build an :class:`~qiskit.circuit.Instruction` object from
    a :class:`~qiskit.circuit.QuantumCircuit`

    This is a specialized version of :func:`~qiskit.converters.circuit_to_instruction`
    to avoid deep copy. This requires a quantum circuit whose parameters are all bound.
    Because this does not take a copy of the input circuit, this assumes that the input
    circuit won't be modified.

    If https://github.com/Qiskit/qiskit-terra/issues/7983 is resolved,
    we can remove this function.

    Args:
        circuit(QuantumCircuit): Input quantum circuit

    Returns:
        An :class:`~qiskit.circuit.Instruction` object
    """
    if len(circuit.qregs) > 1:
        return circuit.to_instruction()

    # len(circuit.qregs) == 1 -> No need to flatten qregs
    inst = Instruction(
        name=circuit.name,
        num_qubits=circuit.num_qubits,
        num_clbits=circuit.num_clbits,
        params=[],
    )
    inst.definition = circuit
    return inst


def _statevector_from_circuit(
    circuit: QuantumCircuit, rng: np.random.Generator | None
) -> Statevector:
    """Generate a statevector from a circuit

    If the input circuit includes any resets for a some subsystem,
    :meth:`.Statevector.reset` behaves in a stochastic way in :meth:`.Statevector.evolve`.
    This function sets a random number generator to be reproducible.

    See :meth:`.Statevector.reset` for details.

    Args:
        circuit: The quantum circuit.
        seed: The random number generator or None.
    """
    sv = Statevector.from_int(0, 2**circuit.num_qubits)
    sv.seed(rng)
    return sv.evolve(bound_circuit_to_instruction(circuit))



# This code is part of Qiskit.
#
# (C) Copyright IBM 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Testing instruction alignment pass."""

from qiskit import QuantumCircuit, pulse
from qiskit.transpiler import PassManager
from qiskit.transpiler.exceptions import TranspilerError
from qiskit.transpiler.passes import ValidatePulseGates
from test import QiskitTestCase  # pylint: disable=wrong-import-order
from qiskit.utils.deprecate_pulse import decorate_test_methods, ignore_pulse_deprecation_warnings


@decorate_test_methods(ignore_pulse_deprecation_warnings)
class TestPulseGateValidation(QiskitTestCase):
    """A test for pulse gate validation pass."""

    def test_invalid_pulse_duration(self):
        """Kill pass manager if invalid pulse gate is found."""

        # this is invalid duration pulse
        # this will cause backend error since this doesn't fit with waveform memory chunk.
        with self.assertWarns(DeprecationWarning):
            custom_gate = pulse.Schedule(name="custom_x_gate")
            custom_gate.insert(
                0, pulse.Play(pulse.Constant(100, 0.1), pulse.DriveChannel(0)), inplace=True
            )

        circuit = QuantumCircuit(1)
        circuit.x(0)
        with self.assertWarns(DeprecationWarning):
            circuit.add_calibration("x", qubits=(0,), schedule=custom_gate)

        with self.assertWarns(DeprecationWarning):
            pm = PassManager(ValidatePulseGates(granularity=16, min_length=64))
        with self.assertRaises(TranspilerError):
            pm.run(circuit)

    def test_short_pulse_duration(self):
        """Kill pass manager if invalid pulse gate is found."""

        # this is invalid duration pulse
        # this will cause backend error since this doesn't fit with waveform memory chunk.
        custom_gate = pulse.Schedule(name="custom_x_gate")
        custom_gate.insert(
            0, pulse.Play(pulse.Constant(32, 0.1), pulse.DriveChannel(0)), inplace=True
        )

        circuit = QuantumCircuit(1)
        circuit.x(0)
        with self.assertWarns(DeprecationWarning):
            circuit.add_calibration("x", qubits=(0,), schedule=custom_gate)

        with self.assertWarns(DeprecationWarning):
            pm = PassManager(ValidatePulseGates(granularity=16, min_length=64))
        with self.assertRaises(TranspilerError):
            pm.run(circuit)

    def test_short_pulse_duration_multiple_pulse(self):
        """Kill pass manager if invalid pulse gate is found."""

        # this is invalid duration pulse
        # however total gate schedule length is 64, which accidentally satisfies the constraints
        # this should fail in the validation
        custom_gate = pulse.Schedule(name="custom_x_gate")
        custom_gate.insert(
            0, pulse.Play(pulse.Constant(32, 0.1), pulse.DriveChannel(0)), inplace=True
        )
        custom_gate.insert(
            32, pulse.Play(pulse.Constant(32, 0.1), pulse.DriveChannel(0)), inplace=True
        )

        circuit = QuantumCircuit(1)
        circuit.x(0)
        with self.assertWarns(DeprecationWarning):
            circuit.add_calibration("x", qubits=(0,), schedule=custom_gate)

        with self.assertWarns(DeprecationWarning):
            pm = PassManager(ValidatePulseGates(granularity=16, min_length=64))
        with self.assertRaises(TranspilerError):
            pm.run(circuit)

    def test_valid_pulse_duration(self):
        """No error raises if valid calibration is provided."""

        # this is valid duration pulse
        custom_gate = pulse.Schedule(name="custom_x_gate")
        custom_gate.insert(
            0, pulse.Play(pulse.Constant(160, 0.1), pulse.DriveChannel(0)), inplace=True
        )

        circuit = QuantumCircuit(1)
        circuit.x(0)
        with self.assertWarns(DeprecationWarning):
            circuit.add_calibration("x", qubits=(0,), schedule=custom_gate)

        # just not raise an error
        with self.assertWarns(DeprecationWarning):
            pm = PassManager(ValidatePulseGates(granularity=16, min_length=64))
        pm.run(circuit)

    def test_no_calibration(self):
        """No error raises if no calibration is added."""

        circuit = QuantumCircuit(1)
        circuit.x(0)

        # just not raise an error
        with self.assertWarns(DeprecationWarning):
            pm = PassManager(ValidatePulseGates(granularity=16, min_length=64))
        pm.run(circuit)



# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2018.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Expand a gate in a circuit using its decomposition rules."""

from __future__ import annotations

from collections.abc import Sequence
from typing import Type
from fnmatch import fnmatch
import warnings

from qiskit.transpiler.basepasses import TransformationPass
from qiskit.transpiler.passes.utils import control_flow
from qiskit.dagcircuit.dagnode import DAGOpNode
from qiskit.dagcircuit.dagcircuit import DAGCircuit
from qiskit.converters.circuit_to_dag import circuit_to_dag
from qiskit.circuit.instruction import Instruction

from ..synthesis import HighLevelSynthesis


class Decompose(TransformationPass):
    """Expand a gate in a circuit using its decomposition rules."""

    def __init__(
        self,
        gates_to_decompose: (
            str | Type[Instruction] | Sequence[str | Type[Instruction]] | None
        ) = None,
        apply_synthesis: bool = False,
    ) -> None:
        """
        Args:
            gates_to_decompose: optional subset of gates to be decomposed,
                identified by gate label, name or type. Defaults to all gates.
            apply_synthesis: If ``True``, run :class:`.HighLevelSynthesis` to synthesize operations
                that do not have a definition attached.
        """
        super().__init__()
        self.gates_to_decompose = gates_to_decompose
        self.apply_synthesis = apply_synthesis

    def run(self, dag: DAGCircuit) -> DAGCircuit:
        """Run the Decompose pass on `dag`.

        Args:
            dag: input dag.

        Returns:
            output dag where ``gate`` was expanded.
        """
        # We might use HLS to synthesize objects that do not have a definition
        hls = HighLevelSynthesis(qubits_initially_zero=False) if self.apply_synthesis else None

        # Walk through the DAG and expand each non-basis node
        for node in dag.op_nodes():
            # Check in self.gates_to_decompose if the operation should be decomposed
            if not self._should_decompose(node):
                continue

            if node.is_control_flow():
                decomposition = control_flow.map_blocks(self.run, node.op)
                dag.substitute_node(node, decomposition, inplace=True)

            elif getattr(node.op, "definition", None) is None:
                # if we try to synthesize, turn the node into a DAGCircuit and run HLS
                if self.apply_synthesis:
                    # note that node_as_dag does not include the condition, which will
                    # be propagated in ``substitute_node_with_dag``
                    node_as_dag = _node_to_dag(node)
                    synthesized = hls.run(node_as_dag)
                    dag.substitute_node_with_dag(node, synthesized, propagate_condition=True)

                # else: no definition and synthesis not enabled, so we do nothing
            else:
                rule = node.op.definition.data

                if (
                    len(rule) == 1
                    and len(node.qargs) == len(rule[0].qubits) == 1  # to preserve gate order
                    and len(node.cargs) == len(rule[0].clbits) == 0
                ):
                    if node.op.definition.global_phase:
                        dag.global_phase += node.op.definition.global_phase
                    dag.substitute_node(node, rule[0].operation, inplace=True)
                else:
                    decomposition = circuit_to_dag(node.op.definition)
                    dag.substitute_node_with_dag(node, decomposition)

        return dag

    def _should_decompose(self, node: DAGOpNode) -> bool:
        """Call a decomposition pass on this circuit to decompose one level (shallow decompose)."""
        if self.gates_to_decompose is None:  # check if no gates given
            return True

        if not isinstance(self.gates_to_decompose, list):
            gates = [self.gates_to_decompose]
        else:
            gates = self.gates_to_decompose

        strings_list = [s for s in gates if isinstance(s, str)]
        gate_type_list = [g for g in gates if isinstance(g, type)]

        if (
            getattr(node.op, "label", None) is not None
            and node.op.label != ""
            and (  # check if label or label wildcard is given
                node.op.label in gates or any(fnmatch(node.op.label, p) for p in strings_list)
            )
        ):
            return True
        elif node.name in gates or any(  # check if name or name wildcard is given
            fnmatch(node.name, p) for p in strings_list
        ):
            return True
        elif any(isinstance(node.op, op) for op in gate_type_list):  # check if Gate type given
            return True
        else:
            return False


def _node_to_dag(node: DAGOpNode) -> DAGCircuit:
    # Control flow is already handled separately, however that does not capture
    # c_if, which we are treating here. We explicitly ignore the condition attribute,
    # which will be handled by ``substitute_node_with_dag``, so we create a copy of the node
    # and set the condition to None. Once ``c_if`` is removed for 2.0, this block can go, too.
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", category=DeprecationWarning)
        if getattr(node.op, "condition", None) is not None:
            op = node.op.copy()
            op.condition = None
            node = DAGOpNode(op, node.qargs, node.cargs)

    # create new dag and apply the operation
    dag = DAGCircuit()
    dag.add_qubits(node.qargs)
    dag.add_clbits(node.cargs)
    dag.apply_operation_back(node.op, node.qargs, node.cargs)

    return dag
